//  Programs for simple distribution problems.

//  Data files:
// 	F	D	A	// Factories, depots, areas, each >= 1
//	Cp[1]	K[1]		// Factory capacity, mfg cost
//	  :	  :
//	Cp[F]	K[F]
//	Cp[1]	H[1]		// Depot capacity, handling cost
//	  :	  :
//	Cp[D]	H[D]
//	T[1,1]	. . .	T[1,D]	// Shipping, factory -> depots
//	  :		  :
//	T[F,1]	. . .	T[F,D]
//	B[1,1]	. . .	B[1,A]	// Shipping, factory -> area
//	  :		  :
//	B[F,1]	. . .	B[F,A]
//	C[1,1]	. . .	C[1,A]	// Shipping, depot -> area
//	  :		  :
//	C[D,1]	. . .	C[D,A]
//	Dmd[1]  . . .	Dmd[A]	// Demands by area
//
//  Begin a new line everywhere shown.  Entries should all be >= 0.


//  Solve distribution problem with F D A as given, hence
//
//		 V  = F*D + (F + D)*A   variables,
//		 Cn = F+2*D + A         constraints.


//  Tableau is
//	                                                           V
//        0|1                F*D|F*D + 1              F*D + (F + D)*A|
//     --+-+--------------------+------------------------------------+
//     0 | |    Costs F -> D    |           Costs F,D -> A           |
//     --+-+--------------------+------------------------------------+
//     1 | |                    :                                    |
//     : | |         Factory    :    Capacity                        |
//     F | |                    :                                    |
//     --+-+--------------------+------------------------------------+
//   F+1 | |                    |                                    |
//     : | |          0         |           Depot Capacity           |
//   F+D | |                    |                                    |
//     --+-+--------------------+------------------------------------+
// F+D+1 | |                    :                                    |
//     : | |           Depot    :    Balance                         |
// F+2*D | |                    :                                    | Ar
//     --+-+--------------------+------------------------------------+ |
//  " +1 | |                    |                                    | v
//     : | |          0         |             Area Demand            |
//  " +A | |                    |                                    |
//     --+-+--------------------+------------------------------------+
//                            Ab ->
//  Variable mapping:
//
//	Fx -> Dy	D*(x - 1) + y
//	Fx -> Az	A*(x - 1) + z + F*D
//	Dy -> Az	A*(F + y - 1) + z + F*D

import java.io.*;

public class Dist
{

// Read next number (int) from data stream

static int nexti(StreamTokenizer data) throws IOException
 { int tok = data.nextToken();  return (int) data.nval; }


// Read next number (float) from data stream

static float nextf(StreamTokenizer data) throws IOException
 { int tok = data.nextToken();  return (float) data.nval; }


// Setup problem from dataf and run it

public static void distribution(String dataf)
 { int F, D, A, V, Cn;	// Problem parameters
   float[][] Tx, Tw;	// Tableau, working copy
   float[]   Sk, Rs,	// Final slacks, residuals
	     Va;	// Final variable values
   int []   r, u, v,	// Types, basic vars, non-basic vars
	    istat;	// Optimization status
   int Ab, Ar;		// 0-th col for area vars, 0-th row for area balance

   istat = new int[3];
   try
    { StreamTokenizer data =
	new StreamTokenizer
	    (new BufferedInputStream(new FileInputStream(dataf)));
      data.parseNumbers();      data.slashSlashComments(true);
      F = nexti(data);  D = nexti(data);  A = nexti(data);
      V = F*D + (F + D)*A;        Cn = F+2*D + A;

      // Create principal arrays
      Tx = new float[Cn+1][V+1];  Tw = new float[Cn+1][V+1];
      Sk = new float[Cn+1];       Rs = new float[Cn+1];
      Va = new float[V+1];
      r  = new int[Cn+1];         u  = new int[Cn+1];
      v  = new int[V+1];

      // Build constraints and objective
      System.out.println("F D A V Cn: "+F+" "+D+" "+A+" "+V+" "+Cn);
      // Tx <- 0 (everywhere)
      for ( int i = 0;  i <= Cn;  i++ )  
       for ( int j = 0;  j <= V;  j++ )  Tx[i][j] = 0.0F;

      Ab = F*D;					// Base for area var's
      Ar = F + 2*D;				// Base for area constraints
      for ( int i = 1;  i <= F;  i++ )		// Factory cap's, costs
       { int m = D*(i - 1),			// 0'th Fi -> D var
	     n = A*(i - 1) + Ab;		// 0'th Fi -> area var
	 float cp = nextf(data),		// Capacity
	      cst = nextf(data);		// Cost
	 Tx[i][0] = cp;				// Capacity constraint
	 for ( int j = m+1;  j <= m+D;  j++ )	//  Factory to depots
	  { Tx[i][j] = 1.0F;  Tx[0][j] = cst; }
	 for ( int j = n+1;  j <= n+A;  j++ )	//  Factory to areas
	  { Tx[i][j] = 1.0F;  Tx[0][j] = cst; }
       }

      for ( int i = 1;  i <= D;  i++ )		// Depot cap's, costs
       { int n = A*(F + i - 1) + Ab,		// 0'th Di -> area var
	     b = F + D + i;			// Balance constraint
	 float cp = nextf(data),		// Capacity
	      cst = nextf(data);		// Handling cost
	 Tx[F+i][0] = cp;			// Depot constraints
	 for ( int j = n+1;  j <= n+A;  j++ )	//  Depot to areas
	  { Tx[F+i][j] = 1.0F;  Tx[F+i+D][j] = 1.0F;	// Cap, balance
	    Tx[0][j] = cst;				// Objective
	  }
	 for ( int j = i;  j <= Ab;  j += D )	// F -> D balance
	   Tx[b][j] = -1.0F;
       }

      for ( int i = 1;  i <= F;  i++ )		// F -> D shipping costs
       { int m = D*(i - 1);
	 for ( int j = 1;  j <= D;  j++ )
	   Tx[0][m+j] += nextf(data);
       }
      for ( int i = 1;  i <= F+D;  i++ )	// F,D -> A shipping costs
       { int n = A*(i - 1) + Ab;		// 0'th F/D -> area var
	 for ( int j = n+1;  j <= n+A;  j++ )
	   Tx[0][j] += nextf(data);
       }

      //  Area demands and coefficients for area balance:
      for ( int i = Ar + 1;  i <= Cn;  i++ )
       { Tx[i][0] = nextf(data);		// Demand
	 for ( int j = Ab + i - Ar;  j <= V;  j += A )
	    Tx[i][j] = 1.0F;
       }

   // Constraint types
   r[0] = 0;						// Objective
   for ( int i = 1;      i <= F+D;  i++ ) r[i] = 1;	// <=
   for ( int i = F+D+1;  i <= Cn;   i++ ) r[i] = 0;	// ==
   prDisTab(Cn, V, Tx);					// Display tableau

   // Copy Tx for checking:
   for ( int i = 0;  i <= Cn;  i++ )
    for ( int j = 0;  j <= V;  j++ ) Tw[i][j] = Tx[i][j];

   // Negate objective (we're minimizing):
   for ( int j = 0;  j <= V;  j++ ) Tx[0][j] = -Tx[0][j];

   // Attempt optimization:
   LpSolve.lpSolve(Cn, V, Tx, r, u, v, istat);

   // Negate objective (we minimized):
   for ( int j = 0;  j <= V;  j++ ) Tx[0][j] = -Tx[0][j];

   // Report results:
   System.out.println("Outcome: "+istat[0]+"   Phase 1 cycles: "+istat[1]
			+"   Phase 2 cycles: "+istat[2]);
   System.out.println("Operating cost: "+Tx[0][0]);
   
   // Retrieve results:
   for ( int j = 0;  j <= V;  j++ )  Va[j] = 0.0F;
   for ( int i = 0;  i <= Cn;  i++ ) Sk[i] = 0.0F;
   for ( int i = 1;  i <= Cn;  i++ )
    { int uv = u[i];
      if (uv > 0) Va[uv] = Tx[i][0]; else
      if (uv < 0) Sk[-uv] = Tx[i][0];
    }
   System.out.println("Factory outputs:");	// Capacity - slack
   for ( int i = 1;  i <= F;  i++ ) System.out.print((Tw[i][0] - Sk[i])+"  ");
   System.out.println();
   System.out.println("Depot throughputs:");	// Capacity - slack
   for ( int i = F+1;  i <= F+D;  i++ )
     System.out.print((Tw[i][0] - Sk[i])+"  ");
   System.out.println();

   // Prepare to check solution against original constraints:
   Va[0] = -1.0F;				// For checking (constants)
   Sk[0] = -Tx[0][0];
   // Adjust negative slacks
   for ( int i = 0;  i <= Cn;  i++ ) if ( r[i] < 0 ) Sk[i] = -Sk[i];
   // Calculate residuals as Sk + Tw.Va
   double maxrs = -1.0E30F, minrs = 1.0E30F;
   for ( int i = 0;  i <= Cn;  i++ )
    { double w = 0.0;
      for ( int j = 0;  j <= V;  j++ )
	 w += ((double) Tw[i][j])*((double) Va[j]);
      w =  (Sk[i] + w);
      Rs[i] = (float)w;
      if ( w > maxrs ) maxrs = w;  if ( w < minrs ) minrs = w;
    }
   System.out.println("Residual range:  "+minrs+"  "+maxrs);
   prSolution(F, D, A, V, Cn, Va);
    }
   catch (IOException ex)
    { System.out.println("Couldn't read data file - "+ex.getMessage()); }

 } // end distribution

// ==========================================================================


// Concise display of distribution tableau.  Doesn't attempt to
// print for large problems.  Prints only integer part of value.

static final int MxCn = 60, MxV = 70;  // Max tableau size for this method

static EasyFormat ef = new EasyFormat();

public static void prDisTab(int Cn, int V, float Tb[][])
 { StringBuffer Tbc = new StringBuffer(71);
   int j;

   if ( (Cn > MxCn) || (V > MxV) )
    { System.out.println("Tableau is too large to print.");
      return;
    }
   System.out.println();  System.out.println("Cost:");
   j = 0;
   for ( int i = 0;  i <= V;  i++ )
    { System.out.print(" "+ef.fFormat(Tb[0][i], 6, 0));
      if ( j == 10 )
       { System.out.println();  j = 1;  System.out.print("       "); }
      else j++;
    }
   System.out.println();  if ( j > 1 ) System.out.println();
   System.out.println("Constraints:");
   for ( int i = 1;  i <= Cn;  i++ )
    { Tbc.setLength(0);
      for ( int u = 1;  u <= V;  u++ )
       { float w = Tb[i][u];
         Tbc.append((w > 0.0F) ? '+' : (w < 0.0F) ? '-' : '0');
       }
     System.out.println(" "+ef.fFormat(Tb[i][0], 6, 0)+" "+Tbc.toString());
    }
   System.out.println();
 }

// ===========================================================================

static final int MxA = 99, MxS = 9;		// Limits on problem size

// Print solution of distribution problem.  Intended for modest problems.

public static void prSolution(int F, int D, int A, int V, int Cn,
							float Va[])
 { StringBuffer Tbc = new StringBuffer(81);

   System.out.println();
   if ( (A > MxA) || (F+D > MxS ) )
    { System.out.println("Solution is too large to print.");  return; }
   else
      System.out.println("Solution:");

   Tbc.setLength(0);  Tbc.append(" ");
   for ( int i = 1;  i <= F;  i++ ) 
    { Tbc.append("      F");  Tbc.append(i); }
   for ( int i = 1;  i <= D;  i++ )
    { Tbc.append("      D");  Tbc.append(i); }
   System.out.println(Tbc.toString());
   for ( int j = 1;  j <= D;  j++ )		// Factory -> Depot[j]
    { System.out.print(" D"+j);
      for ( int u = j;  u <= F*D;  u += D )
        System.out.print(" "+ef.fFormat(Va[u], 7, 1));
      System.out.println();
    }
   System.out.println();
   for ( int j = 1;  j <= A;  j++ )		// F/D -> Area[j]
    { System.out.print(" A"+ef.dFormat(j, 2));
      for ( int u = F*D +j;  u <= V;  u += A )
        System.out.print(" "+ef.fFormat(Va[u], 7, 1));
      System.out.println();
    }
   System.out.println();
 }  // end prSolution

}
