// General system to solve LP problems in standard form.

//
//  Maximize A[0,1:N].X[1:N]  subject to  X[1:N] >= 0 and
//
//	A[i,1:N].X[1:N] ## A[i,0]  for 1 <= i <= M
//
//  where, for each constraint, ## is determined from
//
//		r[i] | ##
//		-----+----
//		 -1  | >=
//		  0  | =
//		 +1  | <=
//
//  It is required that A[1:M,0] >= 0.  If A[0,0] <> 0 this will be incoporated
//  in the objective function.
//
//  Outcome is reported in istat as
//
//  istat[0]:
//
//	0 - optimal solution
//	1 - objective not bounded above
//	2 - optimization (phase 2) failed due to degeneracy
//	3 - infeasible
//	4 - phase 1 failed with unbounded feasibility (not expected)
//	5 - phase 1 failed due to degeneracy
//
//  istat[1], istat[2] have the number of simplex cycles in phase1, phase2.
//
//  The final tableau is returned in A, u, v as
//
//	 LHS  | Const.  Z[v[1]] Z[v[2]]	. . .	Z[v[N]]
//    --------+--------------------------------------------
//    Objctv  =	A[0,0]	A[0,1]	A[0,2]	. . .	A[0,N]
//    Z[u[1]] =	A[1,0]	A[1,1]	A[1,2]	. . .	A[1,N]
//    Z[u[2]] =	A[2,0]	A[2,1]	A[2,2]	. . .	A[2,N]
//       :    :     :       :       :               :
//       :    :     :       :       :               :
//    Z[u[M]] =	A[M,0]	A[M,1]	A[M,2]	. . .	A[M,N]
//
//  where Z[u[i]] is X[u[i]] if U[i] > 0, the slack or artificial variable
//  associated with the original constraint -U[i] if U[i] < 0.  Similarly
//  for Z[v[i]].  The program will fill in u, v as appropriately.

import java.io.*;

public class LpSolve
{

public static float eps = 1.0E-10F;    // Tolerance for 0 test

public static void lpSolve(int M, int N, float A[][],
			   int r[], int u[], int v[], int istat[])
 {
  float C1[] = new float[N+1],	// For objective calculations
	C2[] = new float[N+1];  // The initial objective

  int iout1, icyc1, iout2, icyc2;  // Status from phase1, phase2

  // Reverse coefficients (not objective) depending on constraint
  // and fill in variable numbers for slacks/artificials:
  r[0] = 1;	// For convenience (objective)
  u[0] = 0;
  for ( int i = 1;  i <= M;  i++ )
   { int cty = r[i];	// Constraint type
     u[i] = (cty == 0) ? (-i) - M : -i;	// Marks art's for equations
     if ( cty > -1 )
       for ( int j = 1;  j <= N;  j++ )  A[i][j] = -A[i][j];
     else A[i][0] = -A[i][0];
   }

  // Save objective and clear A[i,0:N]
  for ( int j = 1;  j <= N;  j++ ) { C2[j] = A[0][j];  A[0][j] = 0.0F; }

  // Set up variable codes for problem variables
  for ( int j = 0;  j <= N;  j++ ) v[j] = j;

  // Calculate auxilliary objective
  
  for ( int i = 1;  i <= M;  i++ )
   { int cty = r[i];	// Constraint type
     if ( cty == -1 )
       for ( int j = 0;  j <= N;  j++ ) A[0][j] += A[i][j];
     else if ( cty == 0 )
       for ( int j = 0;  j <= N;  j++ ) A[0][j] -= A[i][j];
   }

  //  System.out.println("Initial A for phase 1:");  Simplex.prflt(M, N, A);

  //  Phase 1.  (istat is longer than needed, but OK)
  Simplex.simplex1(M, N, A, u, v, istat);

  iout1 = istat[0];  icyc1 = istat[1];		// Status
  //  System.out.println("Phase1 status: "+iout1+"  cycles: "+icyc1);
  if ( (iout1 > 0) || (A[0][0] < -eps) )	// Can't continue
   { istat[0] = iout1 + 3;  istat[2] = 0;  return; }

  // Check for artificials and adjust coefficients
  for ( int j = 1;  j <= N;  j++ )
   if ( v[j] < - M ) 
    { v[j] += M;  for ( int i = 0;  i <= M;  i++ ) A[i][j] = 0.0F; }
  for ( int i = 1;  i <= M;  i++ )
   { if ( u[i] < -M )
       { u[i] += M;  A[i][0] = 0.0F;
         for ( int j = 1;  j <= N;  j++ )
          if ( A[i][j] >= eps )        // A rare occurrence
	    for ( int ix = 1;  ix <= M;  ix++ ) A[ix][j] = 0.0F;
	  else
	    A[i][j] = 0.0F;
       }
     else
       if ( A[i][0] < 0.0F ) A[i][0] = 0.0F;   // In case of rounding.
   }

  // Set up original objective in current basis
  A[0][0] = C2[0];
  for ( int j = 1;  j <= N;  j++ )
   { int vv = v[j];  
     if ( vv > 0 ) A[0][j] = C2[vv]; else A[0][j] = 0.0F;
   }
  for ( int i = 1;  i <= M;  i++ )
   { int uu = u[i];
     if ( uu > 0 )		// Add Obj[uu]*A[i][0:M]
      { float cf = C2[uu];
	for ( int j = 0;  j <= N;  j++ ) A[0][j] += cf*A[i][j];
      }
   }

  // System.out.println("Initial A for phase 2:");  Simplex.prflt(M, N, A);

  // Optimize
  Simplex.simplex(M, N, A, u, v, istat);
  
  // Fill in result status.  istat[0] is already correct
  istat[2] = istat[1];  istat[1] = icyc1;
 } // end lpSolve

// ==========================================================================

//  Set up problem description in standard form from data file `dataf'
//  of the form:
//
//	M	N			// Constraints, problem variables
//	x/n				// Maximize or minimize?
//	 A[0,1] . . . A[0,N] B[0]	// Objective
//	###				// Constraint type
//	 A[1,1] . . . A[1,N] B[1]	// Constraint
//	    :     :      :    :
//	    :     :      :    :		// Program requires B[1:M] >= 0
//	###				// Constraint type
//	 A[M,1] . . . A[M,N] B[M]	// Constraint
//
//  Separate entries with white space.
//
//  x/n is one of { max | MAX | x?? | X?? } for maXimize
//      or one of { min | MIN | n?? | N?? } for miNimize
//  ### is one of { =* | <* | >* }  to indicate =, <=, >=
//	? stands for any character, * stands for any 0 or more char's
//
//  The problem is to optimize  A[0,1:N].X[1:N] + B[0] .
//  The program constructs the problem array A and relation array r
//  as described above.  These need not be initialized before calling
//  lpLoad.  At return A has the coefficients and constants, r has
//  relations codes, with r[0] = +1 to maximize, -1 to minimize.
//  A and r are returned in the packet produced by lpLoad.

public static Packet lpLoad(String dataf)
 { int tok,		// token status
       M, N;    	// problem dimensions
   float A[][];		// Coefficients
   int     r[];		// Constraint types
   String typ;    	// problem or constraint type
   char tc0, tc2; 	// characters from typ 
   Packet pkg = new Packet();	// For result

   A = new float[1][1];  r = new int[1];  // So javac doesn't fuss
   try
    { StreamTokenizer data =
	new StreamTokenizer
	    (new BufferedInputStream(new FileInputStream(dataf)));
      data.parseNumbers();      data.slashSlashComments(true);
      data.wordChars('<', '>');  // So < = > count as letters
      tok = data.nextToken();   M = (int) data.nval;
      tok = data.nextToken();   N = (int) data.nval;
      A = new float[M+1][N+1];  r = new int[M+1];
      tok = data.nextToken();   typ = data.sval;
      tc0 = typ.charAt(0);      tc2 = typ.charAt(2);
      r[0] = -1;
      if (    (tc0 == 'x') || (tc0 == 'X')
           || (tc2 == 'x') || (tc2 == 'X') ) r[0] = 1;
      // Read objective:
      for ( int j = 1;  j <= N;  j++ )
       { tok = data.nextToken();  A[0][j] = (float) data.nval; }
      tok = data.nextToken();  A[0][0] = (float) data.nval;

      // Read constraints:
      for ( int i = 1;  i <= M;  i++ )
       { tok = data.nextToken();  typ = data.sval;
	 tc0 = typ.charAt(0);
	 r[i] = (tc0 == '>') ? -1 : (tc0 == '=') ? 0 : 1;
         for ( int j = 1;  j <= N;  j++ )
	  { tok = data.nextToken();  A[i][j] = (float) data.nval; }
	 tok = data.nextToken();  A[i][0] = (float) data.nval;
       }
    }
   catch (IOException ex)
    { System.out.println("Couldn't read data file - "+ex.getMessage()); }
   pkg.A = A;  pkg.r = r;
   return pkg;
 } // end lpLoad

// ==========================================================================

//  Set up and run calculation in data file 'dataf' (as above).

public static void lpSolver(String dataf)
 { int M, N;		// Problem dimensions
   boolean mxm;		// true if maximizing

   float A[][];		// Problem array, solution tableau

   int r[], u[], v[];		// relations, variable codes
   int istat[] = new int[3];	// result status

   Packet p = lpLoad(dataf);		// Load the data
   A = p.A;  r = p.r;
   M = A.length - 1;  N = A[0].length - 1;
   u = new int[M+1];  v = new int[N+1];
   mxm = (r[0] > 0);
   if ( !mxm )		// Negate objective to minimize
     for ( int j = 0;  j <= N;  j++ ) A[0][j] = -A[0][j];

   // Attempt the solution:
   lpSolve(M, N, A, r, u, v, istat);

   if ( !mxm )		// Negate objective if minimizing
     for ( int j = 0;  j <= N;  j++ ) A[0][j] = -A[0][j];
   System.out.println(" Outcome:  "+istat[0]+"   Phase 1 cycles:  "
		      +istat[1]+"   Phase 2 cycles:  "+istat[2]);
   istat[1] += istat[2];   // For report (total cycles)
   if ( istat[0] >= 3 ) istat[0] -= 3;  // In case of failure.
   Simplex.results(M, N, A, u, v, istat);
 }   // end lpSolver

}
