//  Basic routines for simplex method
//  File:  Simplex.java

//  Tableau:
//
//	LHS  |  Const.  X[v[1]]	X[v[2]]	. . .	X[v[N]]
//    --------+--------------------------------------------
//    Objctv  |	A[0,0]	A[0,1]	A[0,2]	. . .	A[0,N]
//    X[u[1]] |	A[1,0]	A[1,1]	A[1,2]	. . .	A[1,N]
//    X[u[2]] |	A[2,0]	A[2,1]	A[2,2]	. . .	A[2,N]
//       :    |     :       :       :               :
//       :    |     :       :       :               :
//    X[u[M]] |	A[M,0]	A[M,1]	A[M,2]	. . .	A[M,N]
//
//  Requires A[1:M,0] >= 0 .
//
//  Each row is read as
//
//	LHS = A[i,0] + A[i,1]*X[v[1]] + A[i,2]*X[v[2]] +...+ A[i,N]*X[v[N]]
//
//  Note that array dimensions are M+1 and N+1.


public class Simplex
{ 

public static int   dlim = 100;		// Max consecutive degeneracies
public static float eps  = 1.0E-20F,	// Tolerance for 0 test
		    inf = 1.0E37F;	// Very large number

//========================================================================

//  Attempt maximization of the objective.  Leaves updated tableau in
//  A, u, v.  Outcome is reported in istat[] as
//
//	istat[0] = 0	Maximal solution found
//	istat[0] = 1	Objective can be arbitrarily large
//	istat[0] = 2	Failed due to unmanageable degeneracy
//
//	istat[1] reports number of cycles.

public static void
   simplex(int M, int N, float A[][], int u[], int v[], int istat[])
 {

  int dcount,	// Counts steps with degenerate update
        itmp,	// For swap
        p, q,	// Pivot coordinates (normal pivot)
      dp, dq,	// Pivot coordinates (degenerate)
	icyc;   // Cycle counter
   float  pv,	// |Pivot coefficient|
       mxobj,	// Maximum objective coefficient
      pr, cn,	// price, constant
     gn, xgn;	// gain, overall maximum
	   
   boolean bnd; // to identify cases with unbounded objective

   dcount = 0;  icyc = 0;
   dp = 0;  dq = 0;  p = 0;  q = 0;   // So Java sees initialization
   istat[0] = 0; 		      // OK so far
   icyc   = 0;
   // System.out.println("Initial A:");  prflt(M, N, A);

//  Repeat until optimum or failure

   mxobj = maxval(A[0], 1, N);		// Maximum price
   while ( (mxobj > 0.0) && (dcount < dlim) )
    { //  Locate pivot
      xgn = -1.0F;
      for ( int j = 1;  j <= N;  j++ )	// Check columns with price > 0
       { int    cp = 0;	// Potential pivot position in current column
	 float  dl,	// Increase for current col. (delta)
	       cdl;	// min increase for current col.
	 pr = A[0][j];
	 if ( pr > 0.0F )
	  { bnd = false;  cdl = inf;
	    for ( int i = 1;  i <= M;  i++ )
	     { cn = A[i][0];  dl = A[i][j];
	       if ( dl < 0.0F )
	        { dl = -cn/dl;  bnd = true;
	          if ( dl < cdl ) { cdl = dl;  cp = i; }
	        }
	     }
	    if ( ! bnd )  // Unbounded objective
	      { istat[0] = 1;  istat[1] = icyc;  return; }
	    gn = pr*cdl;			    // Gain at this pivot
	    if ( gn > xgn ) { p = cp;  q = j;  xgn = gn; }  // better pivot
	    if ( pr == mxobj )	// In case of degeneracy
	     { dp = cp;  dq = j; }
	   }
         }

      // Check for degeneracy
      if ( xgn < eps ) { dcount++;  p = dp;  q = dq; }
                   else dcount = 0;
      pv = (float) Math.abs(A[p][q]);	// | pivot value |

      // System.out.println("Pivot:  ["+p+", "+q+"]   "+pv);

      // Divide pivot row
      for ( int j = 0;  j <= N;  j++ )  A[p][j] /= pv;

      // Update non-pivot rows, column
      for ( int i = 0;    i < p;   i++ )
       { float cf = A[i][q];  // Coefficient in pivot col.
	 for ( int j = 0;   j < q;   j++ )  A[i][j] += cf*A[p][j];
	 for ( int j = q+1; j <= N;  j++ )  A[i][j] += cf*A[p][j];
       }
      for ( int i = p+1;  i <= M;  i++ )
       { float cf = A[i][q];  // Coefficient in pivot col.
	 for ( int j = 0;   j < q;   j++ )  A[i][j] += cf*A[p][j];
	 for ( int j = q+1; j <= N;  j++ )  A[i][j] += cf*A[p][j];
       }

      // Update pivot column and check for constants < 0.0 (from rounding)
      pv = -pv;		// Actual pivot value (< 0.0)
      A[0][q] /= pv;    // Update price in pivot column
      for ( int i = 1;  i <= M;  i++ )
       {  A[i][q] /= pv;
	  if ( A[i][0] < 0.0F ) A[i][0] = 0.0F;
       }
      A[p][q] = 1.0F/pv;  // Correct value in pivot position.
      // Swap variables
      itmp = u[p];  u[p]   = v[q];  v[q] = itmp;
      icyc   = icyc + 1;			// One more cycle
      // System.out.println("Current A");  prflt(M, N, A);
      mxobj = maxval(A[0], 1, N);
    }
   if ( dcount == dlim ) istat[0] = 2;		// Degeneracy failure
   istat[1] = icyc;
}     // End simplex


//===========================================================================

//  Modified simplex method for phase1:
//  Permits negative constants in constraints, representing infeasibilities.
//  Assumes objective includes -(sum of infeasibilities).

//  Attempt maximization of the objective.  Leaves updated tableau in
//  A, u, v.  Outcome is reported in istat[] as for `simplex' above.
//  Note that, although the program adjusts the objective when artificial
//  variables change to negative slack variables, it does not adjust
//  variable codes in u, v.  Programs which use simplex 1 should do
//  this afterwards, following whatever variable coding conventions
//  they wish.

public static void
   simplex1(int M, int N, float A[][], int u[], int v[], int istat[])
 {

  int dcount,	// Counts steps with degenerate update
        itmp,	// For swap
        p, q,	// Pivot coordinates (normal pivot)
      dp, dq,	// Pivot coordinates (degenerate)
	icyc;   // Cycle counter
   float  pv,	// |Pivot coefficient|
       mxobj,	// Maximum objective coefficient
      pr, cn,	// price, constant
     gn, xgn;	// gain, overall maximum
	   
   boolean bnd; // to identify cases with unbounded objective

   dcount = 0;  icyc = 0;
   dp = 0;  dq = 0;  p = 0;  q = 0;   // So Java sees initialization
   istat[0] = 0; 		      // OK so far
   icyc   = 0;
   // System.out.println("Initial A:");  prflt(M, N, A);

//  Repeat until optimum or failure

   mxobj = maxval(A[0], 1, N);		// Maximum price
   while ( (mxobj > 0.0) && (dcount < dlim) )
    { //  Locate pivot
      xgn = -1.0F;
      for ( int j = 1;  j <= N;  j++ )	// Check columns with price > 0
       { int    cp = 0;	// Potential pivot position in current column
	 float  dl,	// Increase for current col. (delta)
	       cdl;	// min increase for current col.
	 pr = A[0][j];
	 if ( pr > 0.0F )
	  { bnd = false;  cdl = inf;
	    for ( int i = 1;  i <= M;  i++ )
	     { cn = A[i][0];  dl = A[i][j];
	       if ( (dl < 0.0F) && (cn >= 0.0F) )  // A[i][0] may be < 0
	        { dl = -cn/dl;  bnd = true;
	          if ( dl < cdl ) { cdl = dl;  cp = i; }
	        }
	     }
	    if ( ! bnd )  // Unbounded objective
	      { istat[0] = 1;  istat[1] = icyc;  return; }
	    gn = pr*cdl;			    // Gain at this pivot
	    // System.out.println("Col. pivot ["+cp+", "+j+"] "+gn);
	    if ( gn > xgn ) { p = cp;  q = j;  xgn = gn; }  // better pivot
	    if ( pr == mxobj )	// In case of degeneracy
	     { dp = cp;  dq = j; }
	   }
         }

      // Check for degeneracy
      if ( xgn < eps ) { dcount++;  p = dp;  q = dq; }
                   else dcount = 0;
      pv = (float) Math.abs(A[p][q]);	// | pivot value |

      //  System.out.println("Pivot:  ["+p+", "+q+"]   "+pv);

      // Divide pivot row
      for ( int j = 0;  j <= N;  j++ )  A[p][j] /= pv;

      // Update Objective:
       { float cf = A[0][q];  // Coefficient in pivot col.
	 for ( int j = 0;   j < q;   j++ )  A[0][j] += cf*A[p][j];
	 for ( int j = q+1; j <= N;  j++ )  A[0][j] += cf*A[p][j];
       }
      pv = -pv;		// Actual pivot value (< 0.0)
      A[0][q] /= pv;    // Update price in pivot column

      // Update non-pivot rows, columns and adjust objective where
      // artificial variables change to negative slack variables
      for ( int i = 1;    i < p;   i++ )
       { float cf = A[i][q],  // Coefficient in pivot col.
	       cu;	      // New A[i][0]
	 cn = A[i][0];        // Constant, possibly < 0
	 A[i][0] = cu = cn + cf*A[p][0];
	 for ( int j = 1;   j < q;   j++ )  A[i][j] += cf*A[p][j];
	 for ( int j = q+1; j <= N;  j++ )  A[i][j] += cf*A[p][j];
	 if ( (cn < 0.0F) && (cu >= 0.0F) )
	  for ( int j = 0;  j <= N;  j++ ) A[0][j] -= A[i][j];
       }
      for ( int i = p+1;  i <= M;  i++ )
       { float cf = A[i][q];  // Coefficient in pivot col.
	 for ( int j = 0;   j < q;   j++ )  A[i][j] += cf*A[p][j];
	 for ( int j = q+1; j <= N;  j++ )  A[i][j] += cf*A[p][j];
       }

      // Update pivot column {pv = pivot coefficient < 0}s
      for ( int i = 1;  i <= M;  i++ )  A[i][q] /= pv;
      A[p][q] = 1.0F/pv;  // Correct value in pivot position.
      // Swap variables
      itmp = u[p];  u[p]   = v[q];  v[q] = itmp;
      icyc   = icyc + 1;			// One more cycle

      // System.out.println("Current A");  prflt(M, N, A);
      mxobj = maxval(A[0], 1, N);
    }
   if ( dcount == dlim ) istat[0] = 2;		// Degeneracy failure
   istat[1] = icyc;
}     // End simplex1



//===========================================================================

//  Some utility routines


//  Return maximum value in V[a::b]

public static float maxval(float V[], int a, int b)
 { float mx = V[a];
   for ( int i = a+1;  i <= b;  i++ )
    { float x = V[i];  if ( x > mx ) mx = x; }
   return mx;
 }

static EasyFormat ef = new EasyFormat();

//  Print float array (e.g., tableau)  A[0:M][0:N]

public static void prflt(int M, int N, float A[][])
 { for ( int i = 0;  i <= M;  i++ )
    { for ( int j = 0;  j <= N;  j++ ) 
        System.out.print(ef.fFormat(A[i][j], 10, 4));
      System.out.println();
    }
 }


//  Report solution

public static void
   results(int M, int N, float A[][], int u[], int v[], int istat[])
 { String sstat[] = {"Optimal solution found.",
		     "Unbounded objective.   ",
		     "Failed from degeneracy." };
   int minmn = (M < N ? M : N);

   System.out.print(sstat[istat[0]]);
   System.out.println("     Cycles:  "+ef.dFormat(istat[1], 5));
   System.out.println("Final state:      Objective = "+A[0][0]);
   System.out.println("       Basic    Value      Non-basic   Price");
   for ( int i = 1;  i <= minmn;  i++ )
    System.out.println(" "+ef.dFormat(u[i], 8)+" "+ef.fFormat(A[i][0], 13, 6)
	              +" "+ef.dFormat(v[i], 8)+" "+ef.fFormat(A[0][i], 13, 6));
   if ( M > N )
    { for ( int i = minmn + 1;  i <= M;  i++ )
       System.out.println(
		" "+ef.dFormat(u[i], 8)+" "+ef.fFormat(A[i][0], 13, 6));
    }
   else
    { for ( int i = minmn + 1;  i <= N;  i++ )
       System.out.println("                    " // 20 spaces
		  +ef.dFormat(u[i], 8)+" "+ef.fFormat(A[i][0], 13, 6));
    }
   return;
 }

}  // end of class Simplex
