//  Elementary demonstrations using D-dimensional harmonic oscillators.

//  Experiments use random initial conditions abs(X) <= 1.0 and conduct
//  trials for various time steps.  D is number of dimensions, N = 2*D.
//  K is the spring constant, the mass is taken to be 1.0 .  The time
//  interval for an experiment is [0:t1]; each calculation uses M*S steps,
//  reporting every S steps.  The state vector is organized as
//
//	x[0] x[1] . . . x[D-1] v[0] v[1] . . . v[D]
//	 0    1          D-1    D   D+1        N-1  where N = 2*D
//
//  The integration is conducted as M groups of S steps, printing a report
//  after each group, but plotting every value calculated.

import java.util.*;
import java.awt.*;

// The simple (undamped) oscillator

public class Harmonic implements Differential, VPlotter
{
 public static float epsilon = 1E-10F;
 public static final float pi2 = (float) (3.14159265/2.0);

static EasyFormat ef = new EasyFormat();

 public Plotter plt;

 // Parameters for multiband plotting:

 float Bht,	// Band height
       Bsc;     // Scaling factor

 // Parameters for differential equation:

 int Dx = 2, Nx = 4;
 float Kx = 1.0F, Lx = 0.0F;


 public void harmonic(int D, float K, float L, float t1, int M, int S,
		      Plotter pl)
  { int N = 2*D;	// Array dimension

    Solver1 solve = new ModEuler();
    
    float[] X0 = new float[N],		// Initial state
	     X = new float[N], 		// Updated state
	     A = new float[N],		// Amplitude/phase
	    X1 = new float[N];		// Final analytic solution

    Random rng = new Random(28596638L);    

    float W;			// Problem parameters
    float maxamp;		// Maximum amplitude (x or v)
    float h, t, hs;		// Integration parameters

    plt = pl;			// Record plotter for use by `plot' below
    Dx = D;  Nx = N;  Kx = K;  Lx = L;
    if ( K < epsilon )
     { K = 1.0F;  System.out.println("Unsuitable K, using "+K); }
    W = (float) Math.sqrt(K);
    
    // Set up problem with random initial state,  -1.0 <= X <= 1.0

    maxamp = 0.0F;			// Max amplitude (x or v)
    for ( int i = 0;  i < D;  i++ )
     { float x = (float)(2.0*rng.nextFloat() - 1.0),	// Position in coord i
             v = (float)(2.0*rng.nextFloat() - 1.0),	// Velocity
	     amp, phi;
       X0[i] = x;  X0[i+D] = v;
       if ( Math.abs(v) < epsilon )		// velocity = 0
	{ A[i] = amp = v/W;  A[D+i] = phi = pi2; }
       else
	{ A[D+i] = phi = (float) Math.atan2(x, v/W);
	  A[i] = amp = (float)(x/Math.sin(phi));
	}
       if ( amp > maxamp ) maxamp = amp;	// Determine maxamp
       X1[i]   =   (float)(amp*Math.sin(W*t1 + phi));	// Analytic solution
       X1[D+i] = (float)(W*amp*Math.cos(W*t1 + phi));	// at t1
     }
    if ( W > 1.0F ) maxamp *= W;
    System.out.println("maxamp: "+maxamp);

    // Plot scaling:

    Bht = (float) (1.0F/N);  Bsc = (float) (Bht/(2.0*maxamp));
    plt.clear();
   
    // Prepare for integration

    X  = X0;
    h  = t1/(M*S);		// Time step
    hs = t1/M;			// Time for S steps
    t  = 0.0F;			// Initial time
    for ( int i = 1;  i <= M;  i++ )
     { X = solve.nstep(this, X, t, S, h, this);
       t = hs*i;
       prState(X, t);
     }
    System.out.println("Analytic solution: (for undamped oscillator)");
    prState(X1, t1);
    float mxerr = 0.0F;
    for ( int j = 0;  j < N;  j++ )
      mxerr = Math.max(mxerr, Math.abs(X[j] - X1[j]));
    System.out.println(" Max abs error: "+mxerr);
  }


 //  The differential equation - see global parameters above.
 //  This works for damped oscillator, but is also correct for
 //  undamped oscillator with Lx = 0.0F.

 public float[] deriv(float X[], float t)
  { float F[] = new float[Nx];	// Result
   
    for ( int i = 0;  i < Dx;  i++ )
     { F[i] = X[Dx+i];  F[Dx+i] = -Kx*X[i] - Lx*X[Dx+i]; }
    return F;
  }


 //  Plot state vector X:

 public void plot(float X[], float t)
  { for ( int i = 0;  i < Dx;  i++ ) // Plot X[i], X[D+i]
     { float yp = (float)((2*i + 0.5)*Bht);
       plt.drawPoint(t, X[i+Dx]*Bsc + yp,    Color.green);
       plt.drawPoint(t, X[i]*Bsc + yp + Bht, Color.blue );
     }
    plt.repaint();
  }


 //  Print state vector X for time t:

 public void prState(float X[], float t)
  { System.out.println("Simulated time:  "+t);
    for ( int j = 0;   j < Dx;  j++ )
       System.out.print( ef.fFormat(X[j], 13, 7));
    System.out.println();
    for ( int j = Dx;  j < Nx;  j++ )
       System.out.print( ef.fFormat(X[j], 13, 7));
    System.out.println();
  }
}
