//  Elementary data-plotting panel with labels.
//  The current version requires axis control from the outside,
//  but it's easy to imagine interaction.  Data are taken
//  as floats, since greater accuracy isn't relevant.
//
//  The coordinate system is standard Cartesian coordinates.
//  Screen coordinates are hidden by the Plotter.
//
//	+--------------------------------------------------+
//  yhi |					    ^	   |
//      |<------------------ pwd pixels ------------|----->|
//	|					    |	   |
//	|					pht pixels |
//	|					    |	   |
//	|					    |	   |
//  ylo	|					    v	   |
//	+--------------------------------------------------+
//	 xlo                                            xhi
//
//  The actual plot is double buffered, both as a convenience for
//  drawing and to reduce flicker.

import java.awt.*;

public class Plotter extends Panel
{
  static final int horizontal = 1, vertical = 2;  // panel types

  BuffCanvas plot;	// The drawing area
  PLabel yLabel;	// Labels for y (vertical) axis
  PLabel xLabel; 	// Labels for x (horizontal) axis
  Graphics pg;          // Graphics for the buffer
  Point    pp;          // For screen coordinates


  Object psafety;	// Synchronization object


  Color bg = Color.white,
        fg = Color.black;  // May be modified

  int   pwd,	// Canvass width
        pht;	// Canvass height

  float xlo = 0.0F, xhi = 1.0F,	// Current plot range
        ylo = 0.0F, yhi = 1.0F,
	dx =  1.0F, dy = -1.0F; // xhi - xlo, ylo - yhi

  public void init(int wd, int ht)	// Dimensions for plot area
   { psafety = new Object();
     pwd = wd;  pht = ht;
     pp = new Point();
     setLayout(new BorderLayout());
     plot = new BuffCanvas();  add("Center", plot);
     plot.init(pwd, pht);       pg = plot.pg;
     plot.setBackground(bg);  plot.setForeground(fg);
     clear();
     yLabel = new PLabel(vertical);
     add("West", yLabel);
     xLabel = new PLabel(horizontal);
     add("South", xLabel);
   }

  // Clear buffer

  public void clear()
   { pg.setColor(bg);  pg.fillRect(0, 0, pwd, pht); }

  public void setXlo(float x)
   { xLabel.setLo(x);  xlo = x;  dx = xhi - x; }

  public void setXhi(float x)
   { xLabel.setHi(x);  xhi = x;  dx = x - xlo; }

  public void setYlo(float y)
   { yLabel.setLo(y);  ylo = y;  dy = y - yhi; }

  public void setYhi(float y)
   { yLabel.setHi(y);  yhi = y;  dy = ylo - y; }

  public void drawPoint(float x, float y, Color c)
   { pg.setColor(c);
     coords(x, y);  pg.drawRect(pp.x, pp.y, 1, 1);
   }

  public void drawBigPoint(float x, float y, Color c)
   { pg.setColor(c);
     coords(x, y);  pg.fillRect(pp.x - 1, pp.y - 1, 4, 4);
   }

  public void drawXaxis()
   { if ( (ylo <= 0.0F) && (0.0F <= yhi) )
      { pg.setColor(fg);  coords(0,0);
	pg.drawLine(0, pp.y, pwd-1, pp.y);
      }
   }

  public void drawYaxis()
   { if ( (xlo <= 0.0F) && (0.0F <= xhi) )
      { pg.setColor(fg);  coords(0.0F,0.0F);
	pg.drawLine(pp.x, 0, pp.x, pht-1);
      }
   }


  // Calculate screen coordinates in pp, adjusting for
  // Cartesian inputs, screen outputs.

  void coords(float x, float y)
   { pp.x = (int) Math.rint(((pwd - 1)*(x - xlo))/dx);
     pp.y = (int) Math.rint(((pht - 1)*(y - yhi))/dy);
   }


  // Double-buffered control

  public void update(Graphics g)
   { paint(g); }

  public void repaint()
   { synchronized (psafety) { paint(getGraphics()); } }

  public void paint(Graphics g)
   {  plot.repaint(); }

   public void destroy()
    { getGraphics().dispose(); }


// This is an inner class, used only by Plotter

class PLabel extends Panel
{
  final int width = 15; // text width

  TextArea hiTxt,	// Displays high value
	   loTxt;	// Displays low value

  float hi, lo;		// The values

  PLabel(int hv)        // hv gives label type (horizontal/vertical)
   { hiTxt = new TextArea(1, width);
     loTxt = new TextArea(1, width);
     if ( hv == vertical)
       { setLayout(new BorderLayout());
	 add("North", hiTxt);  
	 add("Center", new Label("    y"));
	 add("South", loTxt);
       }
     else
       { setLayout(new FlowLayout());
	 add(new Label("     "));
	 add(loTxt);
	 add(new Label("                     x                     "));
	 add(hiTxt);
       }
   }
     
  void setLo(float v)
   { lo = v;  loTxt.setText(Float.toString(v)); }

  void setHi(float v)
   { hi = v;  hiTxt.setText(Float.toString(v)); }

}

}
