/*  4th-order Adams-Bashfort/Adams-Moulton  method for a system of	*/
/*  massive particles.							*/
/*  See NBState.java for data and initialization.			*/
/*  See nbody.notes for the numerical method.				*/
/*					Ernest Sibert, February 1999	*/

public class ABAM4 extends NBState
 {

/* Some additional data for ABAM4: */

  double[][] Ac0, Ac1, Ac2, Ac3, Vz1, Vz2, Vz3;

/*  Data correspond to current time t0 as				*/
/*									*/
/*  Position:				X0	X1			*/
/*  Velocity:	Vz3	Vz2	Vz1	V0	V1			*/
/*  Accel:	Ac3	Ac2	Ac1	Ac0				*/
/*									*/
/*  Time:	 -3	 -2	 -1	 0	 1			*/
/*  Index:	  3	  2	  1	 -

/*  For starting we use RK4.  Introducing a separate state for this	*/
/*  actually doubles the space requirement, but it's easy and these	*/
/*  programs are used only for relatively small test systems.		*/

  static final int ssteps = 3;	/* Number of starting steps needed.	*/
  RK4 rk4 = new RK4();
  int istart = ssteps;		/* Number of starting steps remaining.	*/
  public int nx;		/* RK4 steps/ABAM step during start	*/

/************************************************************************/

/*  Special initialization (overrides but uses NBState.init)		*/

public void init(int nb, double t0)	/* Number of bodies, starting time */
 { super.init(nb, t0);
   Ac0 = new double[NB][3];	Ac1 = new double[NB][3];
   Ac2 = new double[NB][3];	Ac3 = new double[NB][3];
   Vz1 = new double[NB][3];	Vz2 = new double[NB][3];
   Vz3 = new double[NB][3];
   rk4.init(nb, t0);
 }


/************************************************************************/
/*  Install Body b as next particle.  Overrides but uses		*/
/*   NBState.installBody.  Insures that rk4 has correct G.		*/

public void installBody(Body b)
 { super.installBody(b);  rk4.G = G;  rk4.installBody(b); }



/************************************************************************/
/*  Reset state to initial positions and velocities of bodies.		*/
/*  Overrides but uses NBStat.resetSystem.				*/

public void resetSystem(Body bodies[])
 { super.resetSystem(bodies);  rk4.resetSystem(bodies); }


/************************************************************************/
/*  X <- Y  for state arrays {X, V, Ac}					*/

void copyTo(double X[][], double Y[][])
 { for ( int i = 0;  i < Nc;  i++ )
    for ( int u = 0;  u < 3;   u++ ) X[i][u] = Y[i][u];
 }



/************************************************************************/
/*  Correct X1, V1 using the 4th-order Adams-Moulton corrector.		*/
/*  Returns max absolute change in any component of X, V (may be used	*/
/*  to iterate to convergence).						*/

double updateXV()
 { double mx, w;

   /* Calculate acceleration from X* in Ac: */
   Grav(X1);

   /* Calculate corrected state for next step in X1, V1,	*/
   /* also calculating max abs change in any state component:	*/

   mx = 0.0;				/* Maximum change so far  */
   for ( int i = 0;  i < Nc;  i++ )
    for ( int u = 0;  u < 3;   u++ )
     { w = X0[i][u]
	  + deltat*(  9.0* V1[i][u] + 19.0* V0[i][u]
		    - 5.0*Vz1[i][u] +      Vz2[i][u] )/24.0;
       mx = Math.max(mx, Math.abs(X1[i][u] - w));	/* Change */
       X1[i][u] = w;

       w = V0[i][u]
	  + deltat*(  9.0* Ac[i][u] + 19.0*Ac0[i][u]
		    - 5.0*Ac1[i][u] +      Ac2[i][u] )/24.0;
       mx = Math.max(mx, Math.abs(V1[i][u] - w));	/* Change */
       V1[i][u] = w;
     }
   return mx;
 }


/************************************************************************/
/*  Compute X1, V1 for one time step of deltat.  X0, V0 are not altered.*/
/*  Returns number of acceleration calculations.  This version corrects	*/
/*  exactly once, but the program may easily be modified to correct	*/
/*  more than once or to iterate to convergence.			*/

int oneStep()
 { double Tx[][]; /* Only for reference */

   /* Current acceleration to Ac0:  */
   Grav(X0);  Tx = Ac0;  Ac0 = Ac;  Ac = Tx;

   if ( istart == 0 )		/* Predictor-corrector is started */

    { /* Calculate predicted X*, V* in X1, V1: */
      for ( int i = 0;  i < Nc;  i++ )
       for ( int u = 0;  u < 3;   u++ )
        { X1[i][u] = X0[i][u]
	  	    + deltat*(  55.0* V0[i][u] - 59.0*Vz1[i][u]
			      + 37.0*Vz2[i][u] -  9.0*Vz3[i][u])/24.0;
          V1[i][u] = V0[i][u]
	  	    + deltat*(  55.0*Ac0[i][u] - 59.0*Ac1[i][u]
			      + 37.0*Ac2[i][u] -  9.0*Ac3[i][u])/24.0;

        }

      updateXV();		/* Correct once */
      return 2;
    }

   else			/* Use nx RK steps */

    { int iac = rk4.nSteps(nx);		/* Repeats Grav(X0) */
      copyTo(X1, rk4.X0);  copyTo(V1, rk4.V0);
      istart--;
      return (iac + 1);
    }

 }



/************************************************************************/
/*  Carry out ns steps of integration using current parameters and	*/
/*  time step.  Leaves new state in X0, V0, advances `stime'.		*/
/*  Returns total number of acceleration calculations.			*/

public int nSteps(int ns)
 { int ic = 0;  double TX[][];  /* TX is used only for reference */
   fail = 0;
   for ( int k = 0;  k < ns;  k++ )
    { ic += oneStep();
      /* Update history of V's and Ac's, leaving new state in X0, V0:	*/
      TX = X0;  X0 = X1;  X1 = TX;	/* "Swap" X0, X1  */
      TX = Vz3;  Vz3 = Vz2;  Vz2 = Vz1;  Vz1 = V0;   V0 = V1;  V1 = TX;
      TX = Ac3;  Ac3 = Ac2;  Ac2 = Ac1;  Ac1 = Ac0;  Ac0 = TX;
    }
   stime += ns*deltat;
   return ic;
 }


/************************************************************************/
/*  Initialize integration process with step dt, tolerance eps,		*/
/*  Overrides but uses NBState.initInteg.				*/

public void  initInteg(double dt, double eps, int lim)
 { super.initInteg(dt, eps, lim);  rk4.initInteg(dt/nx, eps, lim);
   istart = ssteps;
 }

 }
