/*  Definitions of various systems for N-body calculations.	*/
/*				 Ernest Sibert - March 1997	*/

/*  Concerning units:						*/
/*  The customary units are  au - earth mass - year , which	*/
/*  are used for positions, velocities, masses of bodies.	*/
/*  On the other hand, this is inconvenient for dimensions of	*/
/*  bodies, e.g., diameters, so for these we use km .		*/
/*  The conversion factor is NBSystems.b2s ("body to system")	*/
/*  Except for the default initialization, the programs don't	*/
/*  really depend on using this particular system of units.	*/
/*								*/
/*  Systems provided:						*/
/*	marciniak6	- Marciniak's 6-body example		*/
/*	AA1994		- Astronmical Almanac 1194 Jan 8 0h TDT	*/
/*	AA1994Moon	- Same, with Earth and Moon separate	*/
/*	AA1994Mx2	- AA1994Moon as binary system		*/
/*      AA1994Sats      - AA1994 with several satellites	*/


import java.awt.*;

public class NBSystems
 { public static double G   = 0.0001185684121,	/* Gravitational const.	*/
		        b2s = 149597870.66,	/* x (km) = x (au)*b2s	*/
		      start = 0.0,		/* Starting time	*/
		       days = 365.2564;		/* Days in time unit	*/

  public static String tunit = "yr";		/* The time unit.	*/

  /* Generally,  x (body units) = x (system units) * b2s .	*/
  /* Creation programs (below) may set b2s accordingly.		*/

  /* Marciniak's 6-body system	*/

  public static Body[] marciniak6()
   { Body bodies[] = new Body[6];  Body b;
     b = new Body();  bodies[0] = b;
     b.init("Earth",	  -0.2594970, -0.9809988,  0.0000000,
			   5.9726963, -1.6298815,  0.0000000,
				1.00,   12756.28, Color.blue,   3);
     b = new Body();  bodies[1] = b;
     b.init("Jupiter",	   1.4145586,  4.8663114, -0.0515851,
			  -2.6809127,  0.8989233, -0.0565152,
			      317.84,  142984.00, Color.yellow, 5);
     b = new RingBody();  bodies[2] = b;
     b.init("Saturn",	  -6.8029321,  6.1397428,  0.1652675,
			  -1.4776006, -1.5171660,  0.0851270,
			       95.17,  120536.00, Color.yellow, 5);
     b = new Body();  bodies[3] = b;
     b.init("Uranus",	 -14.1847610,-11.9941410,  0.1396083,
			   0.9138937, -1.1634981, -0.0161611,
			       14.545,   25559.00, Color.blue,   4);
     b = new Body();  bodies[4] = b;
     b.init("Neptune",	  -7.9557686,-29.2149712,  0.7840482,
			   1.0961944, -0.2943595, -0.0193314,
			       17.25,   24764.00, Color.green,  4);
     b = new Body();  bodies[5] = b;
     b.init("Sun",	   0.0000000,  0.0000000,  0.0000000,
		 	   0.0000000,  0.0000000,  0.0000000,
			   332958.00, 1392000.00, Color.white,  7);
     return bodies;
   }



/*  System from the Astronomical Almanac for the year 1994	*/
/*  Nautical Almanac Office					*/
/*  US Naval Observatory					*/
/*  Washington, US Government Printing Office, 1993		*/
/*  ISBN 0 11 886944 2						*/
/*								*/
/*  These data use the mean equator and equinox of J2000.0 .	*/
/*  Values are for JD 244 9360.5 which is 1994 Jan 8 0h TDT,	*/
/*  and are referenced to the solar system barycenter.		*/

  public static Body[] AA1994()
   { Body bodies[] = new Body[10];  Body b;
     b = new Body();  bodies[0] = b;	// Earth+moon
     b.init("Earth", -.295538748,  0.9438451758, -3.687987617e-5,
		    -6.093500454, -1.9188971142, -2.443044129e-5,
				1.0123001, 12756.28, Color.blue,   3);
     b = new Body();  bodies[1] = b;
     b.init("Mercury", 0.1920398521, -0.3890514828, -4.98689915275e-2,
		       7.1975188275,  4.9841844021, -0.53620733551877,
				0.055273597, 2439.7, Color.yellow, 3);
     b = new Body();  bodies[2] = b;
     b.init("Venus", 0.158116952, -0.7039616444, -1.88079856250e-2,
		     7.163597466,  1.5690303690, -0.39210548980432,
				0.8149994933, 6051.8, Color.pink, 3);
     b = new Body();  bodies[3] = b;
     b.init("Mars",  0.308702052, -1.3910058441, -3.688313064e-2,
		     5.184883012,  1.5376407556, -9.529877344e-2,
				0.1074466594, 3397.0,  Color.red, 3);
     b = new Body();  bodies[4] = b;
     b.init("Jupiter", -4.648490148, -2.82310851613,  0.11578730628696,
		        1.397818456, -2.22880523865, -2.20814817730e-2,
				317.893768081, 71492.0, Color.yellow, 5);
     b = new RingBody();  bodies[5] = b;
     b.init("Saturn",   8.590344852, -4.65646820880, -0.260370090268226,
		        0.862285367,  1.78559260684, -6.512151884847e-2,
				95.1818290452, 60268.0, Color.yellow, 5);
     b = new Body();  bodies[6] = b;
     b.init("Uranus",  7.4523958521, -18.158376840331, -0.1640143750208320,
		       1.3200418544,   0.477209054528, -1.5487754508522e-2,
				14.501134059,  25559.0, Color.blue, 4);
     b = new Body();  bodies[7] = b;
     b.init("Neptune", 10.7751358521, -28.1839293511844,  0.3320543099540,
		        1.0652660188,   0.4143183973013, -0.0331150631729,
				17.238585378,  24764.0, Color.green, 4);
     b = new Body();  bodies[8] = b;
     b.init("Pluto", -16.1964941479, -23.889715238630,  7.241213839017,
		       0.9945527450,  -0.794748453233, -0.202357636826,
				0.11098201267,  1151.0, Color.magenta, 3);
     b = new Body();  bodies[9] = b;
     b.init("Sun",  0.00110585207699, 6.28410679537e-3, -4.8514988035e-5,
		   -4.63707467879e-6, 4.30977497681e-6,  1.1829633878e-7,
				332946.038, 1392000.00, Color.white,  7);
     start = 1994.0 + 7.0/days;		/* 8 Jan 1994, 0h TDT */
     return bodies;
   }


//  AA1994 system with separate bodies for earth and moon.  The lunar
//  orbit and phase are only approximately correct for the epoch.

  static double
   mMoon = 0.012300095715868764,
   rEarth[] =
      {1.3850006977559224e-5, 2.7937235882659028e-5, 1.5603963122463092e-6},
   vEarth[] =
      {-.0023496139581929993, .0011648314046610066, 0.0},
   rMoon[] =
      {-.0011260080650991086, -.002271302315689809, -1.2686050160025907e-4},
   vMoon[] =
      {.1910240385496906, -.09470100327416302, 0.0};


  public static Body[] AA1994Moon()
   { Body bodies[] = new Body[11],  mb,  b1994[] = AA1994(), em = b1994[0];
     bodies[0] = em;			// Earth+Moon (adjusted below)
     mb = new Body();  bodies[1] = mb;	// Moon
     mb.init("Moon", em.X[0], em.X[1], em.X[2], em.V[0], em.V[1], em.V[2],
				       mMoon,    3476.0,  Color.yellow, 2);
     em.mass = 1.0;
     for ( int u = 0;  u < 3;  u++ )
      { mb.X[u] += rMoon[u];   mb.V[u] += vMoon[u];
        em.X[u] += rEarth[u];  em.V[u] += vEarth[u];
      }
     for ( int i = 1;  i < 10;  i++ )  bodies[i+1] = b1994[i];
     return bodies;
   }

//  AA1994 doubled to make binary star system with planets:

  static double displ = 35.0,  dvel = 0.47;

  public static Body[] AA1994Mx2()
   { Body bodies[] = new Body[20],  b,  b1994[] = AA1994();

     //  Copy first system to bodies and adjust:
     for ( int i = 0;  i < 10;  i++ )
      { b = b1994[i];  bodies[i] = b;
	b.X[0] -= displ;  b.V[1] -= dvel;  b.name += "1";
      }
     //  Make another system, copy to bodies, and adjust:
     b1994 = AA1994();
     for ( int i = 0;  i < 10;  i++ )
      { b = b1994[i];  bodies[i+10] = b;
	b.X[0] += displ;  b.V[1] += dvel;  b.name += "2";
      }
     bodies[19].bclr = Color.yellow;  /* Second Sun */
     return bodies;
   }

//  Make Satellite for primary pr with mass m, offsets rx, vx [au/yr]
//  mass [earth], diam [km], color, plotdiam [pix]
//  This supposes primaries are much more massive than satellites.

  static Body mkSatellite(String name,  Body pr, double[] rx, double[] vx,
		            double mass, double diam, Color clr, int pdiam)
   { Body b = new Body();
     b.init(name, pr.X[0] + rx[0], pr.X[1] + rx[1], pr.X[2] + rx[2],
		  pr.V[0] + vx[0], pr.V[1] + vx[1], pr.V[2] + vx[2],
					      mass, diam, clr, pdiam);
     return b;
   }

//  AA1994 with many satellites.  Order of bodies is:
//  Earth Moon Mercury Venus {Mars} ... {Pluto}
//  Satellites (if any) follow their primaries.

//  Satellite data:

  static double
   mTitan = 0.022653275312750146, dTitan = 5150.0,
   rTitan[] =
     {1.872331086037799e-4, -0.007307559627056662, -0.0036429951868458942},
   vTitan[] = 
     {1.1739625926661907, 0.0, 0.060336249249824246};

 static double
   mRhea = 4.1880004779874217e-4 , dRhea = 1530.0,
   rRhea[] =
     {.0031890338749977656, -.001398423592673955, -5.350878874677304e-4},
   vRhea[] = 
     {0.7562023428869664, 1.435130374072281, 0.7562023428869663};

static double
   mIo = .014877428346207094, dIo = 3630.0,
   rIo[] =
     {.0020292202887733073, .0019445491480612948, -2.4175722331491818e-4},
   vIo[] = 
     {-1.995172602811226, 2.330099045445162, 1.995172602811226};

static double
   mEuropa =.008010922955649974 , dEuropa = 3138.0,
   rEuropa[] =
     {.0025266603873369325, -.0027766543404873067, -.002454508910559627},
   vEuropa[] = 
     {2.044389353099204, 2.0443893530992034, -.20822290518524902};

static double
   mGanymede =.024795713910345157 , dGanymede = 5262.0,
   rGanymede[] =
     {.0011329182820868481, -.006298543664991134, -.0031942476019670815},
   vGanymede[] = 
     {2.0387858912305665, .744058479778428, -.7440584797784279};


static double
   mCallisto =.017992787273404306 , dCallisto = 4800.0,
   rCallisto[] =
     {.008185562387375249, -.006570325405808187, -.006947079518014761},
   vCallisto[] = 
     {1.0834824531588394, 1.3498438308881802, -4.804210806664988e-17};


static double
   mTitania =5.887460428048779e-4 , dTitania = 1580.0,
   rTitania[] =
     {4.787131568269299e-5, -.001461146031850529, -.0025206046466710683},
   vTitania[] = 
     {-.2573910916459731, .6240006193433613, -.366609527697388};

static double
   mOberon = 5.031893518554007e-4 , dOberon = 1524.0,
   rOberon[] =
     {5.753033862976927e-4, 1.8180572640522278e-4, .00385364466619537},
   vOberon[] = 
     {.20035448119597402, -.633998795148237, 0.};

static double
   mTriton = .003602864344102723 , dTriton = 2706.0,
   rTriton[] =
     {.001519042515095465, 1.5769100942314334e-4, -.0018141928724247718},
   vTriton[] = 
     {-.39475886191790477, -.7388703996518403, -.39475886191790477};



  public static Body[] AA1994Sats()
   { Body bodies[] = new Body[20], pb, b1994[] = AA1994Moon();
     int i;
     for ( i = 0;  i < 4;  i++ ) bodies[i] = b1994[i];
     /* Mars is next */
     bodies[i++] = b1994[4];			// Mars
     
     pb = b1994[5]; bodies[i++] = pb;			  // Jupiter
     bodies[i++] = mkSatellite("Io", pb, rIo, vIo,
			       mIo, dIo, Color.red, 2); 
     bodies[i++] = mkSatellite("Europa", pb, rEuropa, vEuropa,
			       mEuropa, dEuropa, Color.green, 2); 
     bodies[i++] = mkSatellite("Ganymede", pb, rGanymede, vGanymede,
			       mGanymede, dGanymede, Color.magenta, 2); 
     bodies[i++] = mkSatellite("Callisto", pb, rCallisto, vCallisto,
			       mCallisto, dCallisto, Color.blue, 2); 	       
     
     pb = b1994[6];  bodies[i++] = pb;		           // Saturn
     bodies[i++] = mkSatellite("Titan", pb, rTitan, vTitan,
			       mTitan, dTitan, Color.yellow, 3);
     bodies[i++] = mkSatellite("Rhea", pb, rRhea, vRhea,
			       mRhea, dRhea, Color.yellow, 2);
     
     pb = b1994[7];  bodies[i++] = pb;		          // Uranus
     bodies[i++] = mkSatellite("Titania", pb, rTitania, vTitania,
			       mTitania, dTitania, Color.white, 2);
     bodies[i++] = mkSatellite("Oberon", pb, rOberon, vOberon,
			       mOberon, dOberon, Color.white, 2);

     pb = b1994[8];  bodies[i++] = pb;		         // Neptune
     bodies[i++] = mkSatellite("Triton", pb, rTriton, vTriton,
			       mTriton, dTriton, Color.red, 3);

     bodies[i++] = b1994[9];			         // Pluto
     bodies[i++] = b1994[10];			         // Sun 
     return bodies;
   }


  //  Select system according to parm:

  public static Body[] select(String parm)
   { Body wbodies[];
     if ( parm.equalsIgnoreCase("mcnk6") )
      { wbodies = marciniak6(); }
     else if ( parm.equalsIgnoreCase("AA1994") )
      { wbodies = AA1994(); }
     else if ( parm.equalsIgnoreCase("AA1994Moon") )
      { wbodies = AA1994Moon(); }
     else if ( parm.equalsIgnoreCase("AA1994Mx2") )
      { wbodies = AA1994Mx2(); }
     else if ( parm.equalsIgnoreCase("AA1994Sats") )
      { wbodies = AA1994Sats(); }
     else
      { System.out.println("UNKNOWN SYSTEM SELECTED - CAN'T START");
	wbodies = new Body[1];
      }
     return wbodies;
   }
 }
