/*  Greenspan Discrete Mechanics for a system of massive particles.	*/
/*  This records the system state and provides methods for initializing	*/
/*  updating same.							*/

import java.lang.Math;
import java.io.*;
import java.net.*;

public class Greenspan extends NBState
 {

   /* State provided by NBState has everything that's needed here. */


/************************************************************************/
/*  Calculate Greenspan acceleration for current X0, X1, leaving result	*/
/*  in Ac.								*/

void greenAccel()
 { double Dij[],	/* (X0[j] - X0[i]) + (X1[j] - X1[i])	*/
	  r0,		/* || X0[j] - X0[i] ||			*/
	  r1,		/* || X1[j] - X1[i] ||			*/
	  fi,		/* Multiplier for particle i		*/
	  fj,		/* Multiplier for particle j		*/
	  w, v;		/* Miscellaneous			*/
   int i, j, u;		/* Loop counters			*/

   Dij = new double[3];  clearAc();
   /* Loop over pairs of particles */
   for ( i = 0;  i < Nc-1;  i++ )
    for ( j = i+1;  j < Nc;  j++ )
     { r0 = 0.0;  r1 = 0.0;		/* To sum squares of diplacements */
       for ( u = 0;  u < 3;  u++ )
        { w = X0[j][u] - X0[i][u];  r0 += w*w;
	  v = X1[j][u] - X1[i][u];  r1 += v*v;
	  Dij[u] = w + v;
	}
       r0 = Math.sqrt(r0);  r1 = Math.sqrt(r1);  w = r0*r1*(r0 + r1);
       fi = Mg[j]/w;   fj = Mg[i]/w;
       for ( u = 0;  u < 3;  u++ )
	{ Ac[i][u] += fi*Dij[u];  Ac[j][u] -= fj*Dij[u]; }
     }
  }


/************************************************************************/
/*  Update X1, V1 using current Ac and deltat, returns max absolute	*/
/*  change in any component.						*/

double updateXV()
 { double mx, w, hdt = 0.5*deltat;  int i, u;

   mx = 0.0;					/* Maximum change so far  */
   for ( i = 0;  i < Nc;  i++ ) 		/* Loop over particles    */
    for ( u = 0;  u < 3;  u++ )
     { w  = V0[i][u] + deltat * Ac[i][u];	/* New velocity component */
       mx = Math.max(mx, Math.abs(V1[i][u] - w));	/* Change	  */
       V1[i][u] = w;
       w  = X0[i][u] + hdt*(w + V0[i][u]);	/* New position component */
       mx = Math.max(mx, Math.abs(X1[i][u] - w));	/* Change	  */
       X1[i][u] = w;
     }
   return mx;
 }


/************************************************************************/
/*  Compute X1, V1 for one time step of deltat, using current Ac to	*/
/*  guess initial X1, V1 from current X0, V0.	X0, V0 are not altered.	*/
/*  Returns number of iterations required to converge (to tolerance	*/
/*  epsilon).  Stops in any case after `limit' iterations.		*/
/*  Increments `fail' on iterations which fail to converge.		*/

int oneStep()
 { double mx;  int ic;				/* Max change		*/
   mx = updateXV() + 2.0*epsilon;		/* Forces one iteration	*/
   for ( ic = 0;				/* Iteration counter	*/
	 (mx > epsilon) && (ic < limit);
	 ic++ )
    { greenAccel();  mx = updateXV(); }
   if ( mx > epsilon ) fail++;
   return ic;
 }


/************************************************************************/
/*  Carry out ns steps of integration using current parameters.		*/
/*  Ac should already have been calculated, perhaps from the previous	*/
/*  time step.  Leaves new sate in X0, V0, advances `stime'.		*/
/*  Returns total number of iterations, leaves `fail' with number of	*/
/*  failed iterations.							*/

public int nSteps(int ns)
 { int ic = 0;  double TX[][];  /* TX is used only for reference */
   fail = 0;
   for ( int k = 0;  k < ns;  k++ )
    { ic += oneStep();
      TX = X0;  X0 = X1;  X1 = TX;	/* "Swap" X0, X1  */
      TX = V0;  V0 = V1;  V1 = TX;	/*  and   V0, V1  */
    }
   stime += ns*deltat;
   return ic;
 }


/************************************************************************/
/*  Initialize integration process with step dt, tolerance eps,		*/
/*  iteration limit lim.  May be also be used to change parameters	*/
/*  during integration.  Calculate Ac as Newtonian acceleration to	*/
/*  insure that Ac is available to begin subsequent integration.	*/

public void  initInteg(double dt, double eps, int lim)
 { deltat = dt;  epsilon = eps;  limit = lim;  Grav(X0); }


 }
