// An attempt at easy decimal formatting using java.text.

  // General conventions:
  //	width - total field width
  //	frac  - number of digits after the decimal point
  //    When the conversion goes reasonably, the result will have
  //    exactly `width' characters, padded with blanks on the left
  //    as needed, decimal points aligned (also e's for exponents).
  //    All of the formats will thus lead to nice columns of  figures.
  //    In awkward cases a longer string may be returned.

  // Format descriptions:
  //    i  integer-part digit
  //    f  fraction-part digit
  //    0  leading 0 or non-zero digit
  //    e  exponent marker (float)
  //    d  exponent marker (double)
  //    -  minus or blank as appropriate
  //    +  plus or minus as appropriate

package cpsJava;

import java.text.*;

public class EasyFormat
{ DecimalFormat nf = (DecimalFormat) NumberFormat.getNumberInstance();
  boolean used = false;   /* Any numbers formatted yet? */

  StringBuffer  sb = new StringBuffer(40),	// For result
		eb = new StringBuffer(10);      // For exponents

  FieldPosition fp = new FieldPosition(NumberFormat.INTEGER_FIELD);

  // There's no reason for outside programs to call init.

  public void init()
   { nf.setMinimumIntegerDigits(1);  nf.setGroupingUsed(false);
     used = true;
   }

  // Integer with leading blanks as required.  Longer than requested
  // if value is out of expected range.

  public String dFormat(int x, int width)
   { int ix;
     if ( !used ) init();
     nf.setMaximumIntegerDigits(width);
     nf.setDecimalSeparatorAlwaysShown(false);
     sb.setLength(0);  nf.format(x, sb, fp);
     ix = fp.getEndIndex();  sb.setLength(ix);
     while ( ix < width ) { sb.insert(0, ' ');  ix++; }
     return sb.toString();
   }

  // Float  -iiiiii.fff with leading blanks as required.
  // Longer than requested if value is out of anticipated range.
  // Not intended for enormous magnitudes.

  public String fFormat(float x, int width, int frac) 
   { int ix, iwide = width - frac - 1;  // Allows for .
     if ( !used ) init();
     nf.setMinimumIntegerDigits(1);  nf.setMaximumIntegerDigits(iwide);
     nf.setMinimumFractionDigits(frac);  nf.setMaximumFractionDigits(frac);
     nf.setDecimalSeparatorAlwaysShown(true);
     sb.setLength(0);  nf.format(x, sb, fp);
     ix = fp.getEndIndex();
     while ( ix < iwide ) { sb.insert(0, " ");  ix++; }
     return sb.toString();
   }
  
  // Double  -iiiiii.fff with leading blanks as required.
  // Longer than requested if value is out of anticipated range.
  // Not intended for enormous magnitudes.

  public String fFormat(double x, int width, int frac) 
   { int ix, iwide = width - frac - 1;  // Allows for .
     if ( !used ) init();
     nf.setMinimumIntegerDigits(1);  nf.setMaximumIntegerDigits(iwide);
     nf.setMinimumFractionDigits(frac);  nf.setMaximumFractionDigits(frac);
     nf.setDecimalSeparatorAlwaysShown(true);
     sb.setLength(0);  nf.format(x, sb, fp);
     ix = fp.getEndIndex();
     while ( ix < iwide ) { sb.insert(0, " ");  ix++; }
     return sb.toString();
   }
  

  // Float as -i.ffffffe+0d, leading blanks as required.
  // Expects width >= frac + 7.  Always has decimal point and exponent.
  // Result will be Java default string if field is too narrow.

  private int ew, ef, jl, ep, ex;	// Used by eFormat and its helpers
  private String jf;			// Java default decimal rep.

  // ew - desired width,	ef - desired frac,
  // jl - length of jf,		ep - position of '.' in jf
  // ex - position of 'E' in jf

  public String eFormat(float x, int width, int frac)
   { if ( !used ) init();
     jf = Float.toString(x);
     if ( width - frac < 7 ) return jf; 	// Pointless to continue
     ew = width;  ef = frac;  jl = jf.length();
     ex = jf.lastIndexOf('E');			// Is there an E?
     ep = jf.indexOf('.');			// Where is `.'?
     sb.setLength(0);  sb.append(jf);
     if ( ex < 0 ) medFormat(0, 2, 'e');
            else { int ev = Integer.parseInt(jf.substring(ex+1, jl));
	           medFormat(ev, 2, 'e');
	         }
     for ( jl = sb.length();  jl < width;  jl++ )  sb.insert(0, ' ');
     return sb.toString();
   }

  // Double as -i.ffffffe+00d, leading blanks as required.
  // Expects width >= frac + 8.  Always has decimal point and exponent.
  // Result will be Java default string if field is too narrow.

  public String eFormat(double x, int width, int frac)
   { if ( !used ) init();
     jf = Double.toString(x);
     if ( width - frac < 8 ) return jf; 	// Pointless to continue
     ew = width;  ef = frac;  jl = jf.length();
     ex = jf.lastIndexOf('E');			// Is there an E?
     ep = jf.indexOf('.');			// Where is `.'?
     sb.setLength(0);  sb.append(jf);
     if ( ex < 0 ) medFormat(0, 3, 'e');
            else { int ev = Integer.parseInt(jf.substring(ex+1, jl));
	           medFormat(ev, 3, 'e');
	         }
     for ( jl = sb.length();  jl < width;  jl++ )  sb.insert(0, ' ');
     return sb.toString();
   }

  // Convert number in sb (which has no E) to scientific format with
  // ed digits in exponent and separator character ec.  The initial
  // exponent is taken to be ev.

  void medFormat(int ev, int ed, char ec)
   { int  fw, el;
     // System.out.println("ew,ef,jl,ep,ex: "+ew+"  "+ef+"  "+jl
     //			+"  "+ep+"  "+ex+"  |"+sb.toString());
     if ( sb.charAt(0) != '-' ) { sb.insert(0, ' ');  ep++;  jl++; }
     if  ( ep < 2 ) return;		// An anomaly
     while ( ep > 2 )
      { sb.setCharAt(ep, sb.charAt(ep-1));  ep--;  ev++; }
     sb.setCharAt(2, '.');
     // System.out.println("Sb: |"+sb.toString());
     // Now we have -i.f...f, but perhaps wrong number of f's
     fw = jl - 3;   // Characters in current fraction.
     // System.out.println("fw: "+fw);
     if ( fw > ef ) ev += round(fw, jl);
     else while ( fw < ef ) { sb.append('0');  fw++; }
     eb.setLength(0);  eb.append(Integer.toString(ev));
     // System.out.println("initial eb |"+eb.toString());
     if ( eb.charAt(0) != '-' )  eb.insert(0, '+');
     eb.insert(0, ec);
     ex = eb.length();  el = ed + 2;
     while ( ex < el ) { eb.insert(2, '0'); ex++; }
     // System.out.println("final eb |"+eb.toString());
     sb.append(eb.toString());
   }

  // Round fraction of current width fw > ef to ef digits.
  // fl is length of current -i.ff...f in sb (ignoring exponent).
  // Shortens sb to reflect result.
  // Returns increase in exponent (0 or 1).
  // Allows for ef == 0, which is unlikely but permitted.

  int round(int fw, int fl)
   { int ix = ef+4, iz = '0', dc, dl;
     boolean carry = false;
     //* System.out.println("fw,fl,ix,iz: "+fw+"  "+fl+"  "+ix+"  "+iz);
     while ( fl > ix )		// Look for non-zero digits
       { fl--;  iz = Math.max(iz, sb.charAt(fl)); }
     dc = sb.charAt(--ix);	// First digit discarded
     ix = (ef > 1) ? ix-1 : 1;	// Position of last character kept
     dl = sb.charAt(ix);	// Last digit kept
     //* System.out.println("dc,ix,dl,iz: "+dc+"  "+ix+"  "+dl+"  "+iz);
     if (   (dc > '5')
	 || ( (dc == '5') && ( (iz > '0') || ((dl&1) == 1) ) ) )
      { // round up, ix is index of character to adjust.
        do { dl++;  carry = true;
             if ( dl <= '9' )
	      { sb.setCharAt(ix, (char)dl);  carry = false; }
	     else
	      { sb.setCharAt(ix, (ix > 1) ? '0' : '1');
		ix--;
		if ( ix == 2 ) ix = 1;
		dl = sb.charAt(ix);
	      }
	    }
	 while ( carry && ( ix > 0 ) );
       }
      sb.setLength(ef+3);
      return ( carry ? 1 : 0 );
   }

  public void describe()
   { System.out.println("Grouping used.. "+nf.isGroupingUsed());
     System.out.println("Grouping size.. "+nf.getGroupingSize());
     System.out.println("Min fraction... "+nf.getMinimumFractionDigits());
     System.out.println("Max fraction... "+nf.getMaximumFractionDigits());
     System.out.println("Min integer.... "+nf.getMinimumIntegerDigits());
     System.out.println("Max integer.... "+nf.getMaximumIntegerDigits());
     System.out.println("Multiplier..... "+nf.getMultiplier());
     System.out.println("Negative prefix "+nf.getNegativePrefix());
     System.out.println("Negative suffix "+nf.getNegativeSuffix());
     System.out.println("Positive prefix "+nf.getPositivePrefix());
     System.out.println("Positive suffix "+nf.getPositiveSuffix());
     System.out.println("Dec sep always. "
			       +nf.isDecimalSeparatorAlwaysShown());
     System.out.println("Pattern........ "+nf.toPattern());
     System.out.println("Examples:");
     System.out.println( -1.2345678E13+"  "
			+ 1582+"  "
			+ -1582+"  "
			+ 0.12345+"  "
			+ 1.2345E-14);
     System.out.println( nf.format(-1.2345678E13)+"  "
			+nf.format(1582)+"  "
			+nf.format(-1582)+"  "
			+nf.format(0.12345)+"  "
			+nf.format(1.2345E-14));
   }

}

