/*  Basic solar system simulation with ViewGroups		*/
/*  implementing multiple views.				*/
/*  Uses 4th-order Runge Kutta as implented in RK4.java		*/
/*				Ernest Sibert - February 1999	*/

import java.awt.*;

public class NBApplt extends java.applet.Applet implements Runnable
 {
   int  N;		/* Number of bodies	*/

   Body bodies[];	/* The bodies 		*/

   ViewGroup viewxyz;	/* Solar system - xy, zx planes	*/

   NBState state;	/* Current state, integration	*/

   /*  Applet control items **********************************	*/

   Thread NBThread;		/* Runs the applet		*/

   Object safety;		/* For synchronization		*/

   int delay = 2000;		/* To slow it down (msec.)	*/

   boolean ready = false,	/* To synchronize startup	*/
	 ptrails = false;	/* Draw permanent trails	*/

   String parm;			/* For applet parameters	*/

   int ns = 10,		/* Number of steps between plots	*/
       np = 10,		/* Number of times to plot		*/
       gs =  0,		/* Loop counter for plotting		*/
     ilim = 10;		/* Iteration limit			*/

   double eps = 1.0e-8,	/* Convergence criterion	*/
	   dt = 0.01,	/* Time step			*/
	width = 80.0;	/* Initial view width		*/


   /*  Initialize	*/

   public void init()
    {  safety = new Object();
       parm = getParameter("nbsys");
       if ( parm != null)
	{ if ( parm.equalsIgnoreCase("mcnk6") )
	  bodies = NBSystems.marciniak6();
        }
       N = bodies.length;
       System.out.println(parm+" "+N);
       setLayout(new FlowLayout(FlowLayout.CENTER));
       viewxyz = new ViewGroup();  add(viewxyz);
       viewxyz.init(2, 500, 500, 250, N);
       viewxyz.addView(1, 2,  3);	/* Standard x-y view	*/
       viewxyz.addView(1, 3, -2);	/* x-z viewed from -y	*/
       state = new RK4();  state.init(N, 0.0);
       installSystem();
       viewxyz.setControls();
       parm = getParameter("vwidth");
       if ( parm != null ) width = Double.valueOf(parm).doubleValue();
       viewxyz.setView(width, 0.0, 0.0, 0.0);
       System.out.println(
	  "au->pix: "+viewxyz.scale+"  km->pix: "+viewxyz.bscale);
    /*  Integration parameters */
       parm = getParameter("deltat");
       if ( parm != null ) dt = Double.valueOf(parm).doubleValue();
       parm = getParameter("epsilon");
       if ( parm != null ) eps = Double.valueOf(parm).doubleValue();
       parm = getParameter("limit");
       if ( parm != null ) ilim = Integer.valueOf(parm).intValue();
       state.initInteg(dt, eps, ilim);
    /*  Integration time, plot controls */
       parm = getParameter("nsteps");
       if ( parm != null ) ns = Integer.valueOf(parm).intValue();
       parm = getParameter("nplots");
       if ( parm != null ) np = Integer.valueOf(parm).intValue();
       parm = getParameter("delay");
       if ( parm != null ) delay = Integer.valueOf(parm).intValue();
       parm = getParameter("ptrails");
       if ( parm != null ) ptrails = parm.equalsIgnoreCase("yes");
       viewxyz.setTrails(ptrails);
       ready = true;
    }



   /*  Draw complete system (from scratch) using positions in Xs.	*/

   void drawSystem(double Xs[][])
    { Body b;
      synchronized (safety)
       { viewxyz.drawAll(Xs); }
    }


   /*  Install bodies in state, initialize G.				*/

   void installSystem()
    { NBObj ob;		/* Ignored for the moment	*/
      state.G = NBSystems.G;
      for ( int i = 0;  i < N;  i++ )
       { state.installBody( bodies[i]);
	 ob = viewxyz.loadObject(bodies[i]); 
       }
      
    }



   /*  Move bodies to current positions and replot - preliminary.	*/

   void moveSystem(double Xs[][])
    { viewxyz.drawAll(Xs);
    }


   /* Delay for specified time (in msec.) */

   void pause(int delay)
    { if ( delay > 0 )
       {  try { Thread.sleep(delay); }
          catch (InterruptedException ex) { }
       }
    }

   /* Start process	*/

   public void start()
    { while ( !ready  ) pause(200);		/* Wait for init */
      drawSystem(state.X0);
      if ( NBThread == null )
       { NBThread = new Thread(this);
	 NBThread.start(); }
    }


   /* Stop process  */

   public void stop()
    { if ( NBThread != null )
       synchronized (safety)
	{ NBThread.stop();  NBThread = null; }
    }



   /*  Run system ns*np steps, plotting every ns steps.	*/

   public void run()
    { viewxyz.repaint();
      if ( delay > 0) pause(delay);
      for ( gs = 1;  gs <= np;  )
       { int nac = advance(); drawSystem(state.X0);
	 if ( delay > 0 ) pause(delay);  viewxyz.repaint();
	 System.out.println("Time: "+state.stime
			   +"   Accelerations: "+nac
			   +"   Failures: "+state.fail);
	 state.printState(state.X0, state.V0);
       }
      System.out.println("Finished.");
    }


   /*  Advance over next group of time steps.  Returns total number	*/
   /*  of acceleration calculations (depends on integrator).		*/

   int advance()
    { int na;
      synchronized (safety) { na = state.nSteps(ns);  gs++; }
      return na;
    }

 }


