/*  4th-order Runge-Kutta  method for a system of massive particles.	*/
/*  See NBState.java for data and initialization.			*/
/*					Ernest Sibert, February 1999	*/

/*  In the notation of nbody.notes, the method may be described as:	*/
/*									*/
/*	Vz1 = V(i);		{ Not really copied }			*/
/*	Ac1 = Grav( X(i) );						*/
/*									*/
/*	Vz2 = V(i) + (deltat/2)*Ac1;					*/
/*	Ac2 = Grav( X(i) + (deltat/2)*V(i) );				*/
/*									*/
/*	Vz3 = V(i) + (deltat/2)*Ac2;					*/
/*	Ac3 = Grav( X(i) + (deltat/2)*Vz2 );				*/
/*									*/
/*	Vz4 = V(i) + deltat*Ac3;					*/
/*	Ac4 = Grav ( X(i) + deltat*Vz3 );	{ Kept in Ac }		*/
/*									*/
/*	X(i+1) = X(i) + deltat*{ V(i) + 2*Vz2 + 2*Vz3 + Vz4 }/6;	*/
/*	V(i+1) = V(i) + deltat*{  Ac1 + 2*Ac2 + 2*Ac3 + Ac4 }/6;	*/


public class RK4 extends NBState
 {

/* Some additional data for RK4: */

  double[][] Ac1, Ac2, Ac3, Vz2, Vz3, Vz4;

/************************************************************************/

/*  Special initialization (overrides but uses NBStat.init)		*/

public void init(int nb, double t0)	/* Number of bodies, starting time */
 { super.init(nb, t0);
   Ac1 = new double[NB][3];	Ac2 = new double[NB][3];
   Ac3 = new double[NB][3];
   Vz2 = new double[NB][3];	Vz3 = new double[NB][3];
   Vz4 = new double[NB][3];
 }

/************************************************************************/
/*  Compute X1, V1 for one time step of deltat.  X0, V0 are not altered.*/
/*  Returns number of acceleration calculations.			*/

int oneStep()
 { double hdt = 0.5*deltat;  double Tx[][]; /* Only for reference */

   /* Current acceleration to Ac1:  */
   Grav(X0);  Tx = Ac1;  Ac1 = Ac;  Ac = Tx;

   /* Calculate mid-step X, V in X1, Vz2: */
   for ( int i = 0;  i < Nc;  i++ )
    for ( int u = 0;  u < 3;   u++ )
     {  X1[i][u] = X0[i][u] + hdt* V0[i][u];
       Vz2[i][u] = V0[i][u] + hdt*Ac1[i][u];
     }
   /* Mid-step acceleration to Ac2: */
   Grav(X1);  Tx = Ac2;  Ac2 = Ac;  Ac = Tx;

   /* Calculate mid-step X, V in X1, Vz3: */
   for ( int i = 0;  i < Nc;  i++ )
    for ( int u = 0;  u < 3;   u++ )
     {  X1[i][u] = X0[i][u] + hdt*Vz2[i][u];
       Vz3[i][u] = V0[i][u] + hdt*Ac2[i][u];
     }
   /* Mid-step acceleration to Ac3: */
   Grav(X1);  Tx = Ac3;  Ac3 = Ac;  Ac = Tx;

   /* Calculate full-step X, V in X1, Vz4: */
   for ( int i = 0;  i < Nc;  i++ )
    for ( int u = 0;  u < 3;   u++ )
     {  X1[i][u] = X0[i][u] + deltat*Vz3[i][u];
       Vz4[i][u] = V0[i][u] + deltat*Ac3[i][u];
     }
   /* Full-step acceleration to Ac: */
   Grav(X1);

   /* Calculate state for next step in X1, V1: */
   for ( int i = 0;  i < Nc;  i++ )
    for ( int u = 0;  u < 3;   u++ )
     { double wv, wa;
       wv = Vz2[i][u] + Vz3[i][u];
       X1[i][u] = X0[i][u] + deltat*(  V0[i][u] + wv + wv + Vz4[i][u] )/6.0;
       wa = Ac2[i][u] + Ac3[i][u];
       V1[i][u] = V0[i][u] + deltat*( Ac1[i][u] + wa + wa +  Ac[i][u] )/6.0;
     }

   return 4;
 }


/************************************************************************/
/*  Carry out ns steps of integration using current parameters.		*/
/*  time step.  Leaves new sate in X0, V0, advances `stime'.		*/
/*  Returns total number of acceleration calculations.			*/

public int nSteps(int ns)
 { int ic = 0;  double TX[][];  /* TX is used only for reference */
   fail = 0;
   for ( int k = 0;  k < ns;  k++ )
    { ic += oneStep();
      TX = X0;  X0 = X1;  X1 = TX;	/* "Swap" X0, X1  */
      TX = V0;  V0 = V1;  V1 = TX;	/*  and   V0, V1  */
    }
   stime += ns*deltat;
   return ic;
 }


 }
