;;  Transforming Astronomical Almanac data into more convenient form
;;  Involves translation to barycenter, rotation so plane of earth's
;;  orbit is x-y plane of coordinate system.


(define pi 3.14159265358979323846)

(define rad (/ 180 pi))          ;; 1 radian = rad degrees

(define yr 365.2564)             ;; days per sidereal year


;; Obliquity of the ecliptic, J 2000, in degrees, cos and sin

(define obliq (+ 23.0 (/ (+ 26.0 (/ 21.4119 60)) 60)))

(define cobq (cos (/ obliq rad)))      ;; (cos obliq)
(define sobq (sin (/ obliq rad)))      ;; (sin obliq)


;;  Vectors are represented as lists of reals,  arrays as lists of vectors.



;;  Some data-parallel operations


;;  Combine list of vectors with op

(define (lmap op L)
  (define (w rs tl)    ;; rs is result so far, tl is tail of L
    (if (pair? tl) (w (map op rs (car tl)) (cdr tl)) rs))
  (if (pair? L) (w (car L) (cdr L)) L)
)


;;  Vector +

(define (v+ . x) (lmap + x))


;;  Vector -

(define (v- . x)
  (let ((ln (length x)))
    (cond ((= ln 1) (map - (car x)))
	  ((= ln 2) (map - (car x) (cadr x)))
	  ((> ln 2) (map - (car x) (lmap + (cdr x)))))
    )
)

;;  Vector *

(define (v* . x) (lmap * x))


;;  Scalar times vector

(define (s*v s v) (map (lambda (x) (* s x)) v))


;;  Vector entries

(define vx car)
(define vy cadr)
(define vz caddr)


;;  Cross product of vectors (a vector)

(define (cross x y)
  (list (- (* (vy x) (vz y)) (* (vz x) (vy y)))
	(- (* (vz x) (vx y)) (* (vx x) (vz y)))
	(- (* (vx x) (vy y)) (* (vy x) (vx y))))
)

;;  Sum reduction of vector

(define (sum x)
  (define (w r tl)        ;; r is sum so far, tl is a tail of x
    (if (pair? tl) (w (+ r (car tl)) (cdr tl)) r))
  (w 0.0 x)
)


;;  Euclidean length of vector

(define (vln x)
  (sqrt (sum (v* x x)))
)


;;  Transformations:  u(helio) + deltau = u(bary)
;;  Note:  These were calculated from the values below
;;         to place barycenter at origin with velocity 0.

(define deltax '( 0.0011058520769863366
		  0.0057848538445339015
		  0.0024551614849180926))

(define deltav '(-4.6370746787943920e-6
		  3.9070859706088060e-6
		  1.8228641180148125e-6))


;;  Transform U = (X V) from heliocentric coordinates [A, A/D]
;;  to barycentric coordinates [A, A/Y], rotated to ecliptic.

(define (beclip U)
  (let ((x (car U)) (v (s*v yr (cadr U))))
    (list (rotx (v+ x deltax) cobq sobq)
          (rotx (v+ v deltav) cobq sobq )))
)


;;  Rotate vector U about x axis by angle with cos c, sin s

(define (rotx U c s)
  (let ((y (cadr U)) (z (caddr U)))
    (list (car U)
	  (+ (* c y) (* s z))
	  (- (* c z) (* s y))))
)

;;  Rotate vector U about y axis by angle with cos c, sin s
(define (roty U c s)
  (let ((x (car U)) (z (caddr U)))
    (list (+ (* c x) (* s z))
          (cadr U)
	  (- (* c z) (* s x))))
	  
)
;;  Rotate vector U about z axis by angle with cos c, sin s

(define (rotz U c s)
  (let ((x (car U)) (y (cadr U)))
    (list (+ (* c x) (* s y))
	  (- (* c y) (* s x))
	  (caddr U)))
)

;;  Transform U = (X V) and print on two lines with commas

(define (pbe U)
  (let ((w (beclip U)))
    (let ((x (car w)) (v (cadr w)))
      (display (car x))   (display ", ")
      (display (cadr x))  (display ", ")
      (display (caddr x)) (display ", ")
      (newline)
      (display (car v))   (display ", ")
      (display (cadr v))  (display ", ")
      (display (caddr v)) (display ", ")
      (newline)))
)

;;  Data for 1994 Jan 8 0h TDT ********************************************

;; Earth, barycentric coords

(define ebx '(-0.295527516  0.865990723  0.375415440))
(define ebv '(-6.097430359 -1.757639730 -0.762277688))


;;  Heliocentric coordinates (X V), velocities are A/D
;;  Mass ratios from 1994 AA, see below for conversions


;;  Earth + moon

(define em '((-0.2966446   0.8601909   0.3729509)
	     (-0.01668279 -0.00482003 -0.00208982)))
(define mem 328900.53)

;;  Sun

(define sun '((0.0 0.0 0.0)
	      (0.0 0.0 0.0)))
(define msun 1.0)

;;  Mercury

(define merc '((0.1909340  -0.3428959  -0.2029648 )
	       (0.01970540  0.01279589  0.00479088)))
(define mmerc 6023600.0)

;;  Venus

(define venus '((0.1570111  -0.6441757  -0.2997309 )
		(0.01961253  0.00436823  0.00072380)))
(define mvenus 408523.0)

;;  Mars

(define mars '((0.3075962  -1.2673366  -0.5896049 )
	       (0.01419520  0.00396615  0.00143516)))
(define mmars 3098710.0)

;;  Jupiter

(define jup '((-4.649596    -2.641994    -1.019190   )
	      ( 0.003826964 -0.005574466 -0.002482718)))
(define mjup 1047.350)

;; Saturn

(define sat '((8.589239    -4.174442   -2.093576   )
	      (0.002360780 0.004556429  0.001780260)))
(define msat 3498.0)

;; Uranus

(define ura '((7.45129     -16.60053     -7.37592    )
	      (0.003614027   0.001215399  0.000481138)))
(define mura 22960.0)

;; Neptune

(define nep '((10.77403    -25.99612    -10.90872    )
	      (0.002916501  0.001076773   0.000368021)))
(define mnep 19314.0)

;; Pluto

(define plu '((-16.19760     -24.80456     -2.86155    )
	      (  0.002722902  -0.001775953 -0.001373813)))
(define mplu 3000000.0)    ;; AA rather than ephemeris


;;  Mass conversions

(define ms/e 332946.038)      ;; (mass of sun)/(mass of earth)


;;  Convert  mr = (mass sun)/(mass planet) to (mass planet)/(mass earth)

(define (mec mr)
  (/ ms/e mr)
)

;; *********************************************************************

;; An approximate earth-moon system, using data from Cambridge
;; Encyclopedia of Astronomy

(define emm (mec mem))		;; Mass Earth+Moon (Earth = 1) from AA1994
(define me 1.0)                 ;; Mass of Earth
(define mm (- emm me))		;; Mass of Moon

(define dist-em 384390.0)	;; semi-major axis (km)

(define per-em 27.3217)		;; Orbital period (d) [sidereal]

(define b2s 149597870.66)	;; km/au

(define e-cm 0.0)		;; Earth to center of mass  (au)
(define m-cm 0.0)		;; Moon  to center of mass  (au)
(define e-spd 0.0)		;; Earth speed relative to c-of-m
(define m-spd 0.0)		;; Moon  speed relative to c-of-m
(define mx (list m-cm 0.0 0.0 ))
(define mv (list 0.0 m-spd 0.0))
(define ex (list e-cm 0.0 0.0 ))
(define ev (list 0.0 e-spd 0.0))

;; Calculate relation to c-of-m and display.

(define (em-orbit)
  (pp "Relation to center of mass: (units are earth, au, yr)")
  (display "Mass of Earth:  ") (display me) (newline)
  (display "Mass of Moon:  ") (display mm) (newline)
  (set! e-cm (/ (* (/ mm emm) dist-em) b2s))
  (display "Earth distance:  ") (display e-cm) (newline)
  (set! m-cm (/ (* (/ emm) dist-em) b2s))
  (display "Moon distance:  ") (display m-cm) (newline)
  (set! e-spd (/ (* 2.0 pi e-cm yr) per-em))
  (display "Earth speed:  ") (display e-spd) (newline)
  (set! m-spd (/ (* 2.0 pi m-cm yr) per-em))
  (display "Moon speed:  ") (display m-spd) (newline)
)

(define m-lambda 236.73) ;; Ecliptic longitude for 1994 Jan 8 0h TDT

(define m-beta -0.050)   ;; Ecliptic latitude for 1994 Jan 8 0h TDT

(define (em-rel)
 (let ((cl (cos m-lambda))  (sl (sin m-lambda))
       (cb (cos m-beta))    (sb  (sin m-beta)))
   (set! mx (list m-cm 0.0 0.0 ))
   (set! mv (list 0.0 m-spd 0.0))
   (set! mx (rotz (roty mx cb (- sb)) cl (- sl)))
   (set! mv (rotz (roty mv cb (- sb)) cl (- sl)))
   (set! ex (s*v (- (/ e-cm m-cm)) mx))
   (set! ev (s*v (- (/ e-spd m-spd)) mv))
  (pp mx)
  (pp mv)
  (pp ex)
  (pp ev)
))

;; > (em-rel)
;; (-.0011260080650991086 -.002271302315689809 -1.2686050160025907e-4)
;; (.1910240385496906 -.09470100327416302 0.)
;; (1.3850006977559224e-5 2.7937235882659028e-5 1.5603963122463092e-6)
;; (-.0023496139581929993 .0011648314046610066 -0.)


;; Calculate barycenter and velocity from lists 
;;
;;    coords:  ((X V) (X V) ... (X V))
;;    masses:  (  m     m   ...   m  )
;;
;; Resturns result and deposits in baryrslt as well.

(define baryrslt '((0.0 0.0 0.0) (0.0 0.0 0.0)))

(define (baryxv coords masses)
  (let ((xvsum (+m*xv coords masses))
	(rsum  (/ (+l masses))))
    (set! baryrslt (list (s*v rsum (car xvsum)) (s*v rsum (cadr xvsum))))
    baryrslt)
)

;; Calculate ( [sum m*x] [sum m*v]) from lists as above.


(define (+m*xv coords masses)
  (define (h xs vs cs ms)
    (if (null? cs) (list xs vs)
        (let ((x (caar cs)) (v (cadar cs)) (m (car ms)))
	  (h (v+ xs (s*v m x)) (v+ vs (s*v m v)) (cdr cs) (cdr ms)))))
  (h '(0.0 0.0 0.0) '(0.0 0.0 0.0) coords masses)
)

;; Calculate sum of list of numbers:

(define (+l nl)
  (if (null? nl) 0.0 (+ (car nl) (+l (cdr nl))))
)
