/*  View with canvas for solar system plotting.			*/
/*  This contains no synchronization.				*/
/*								*/
/*  Provides area for u-v plot understood as	+---> h		*/
/*						|		*/
/*      u	[u, v]  (h, v)		       \|/		*/
/*      ^					v		*/
/*	|						 	*/
/*	+---------------------------------------+		*/
/*	|					|		*/
/*	|<------------ pwd (pixels) ----------->|		*/
/*	|					|		*/
/*	|					|		*/
/*	|					|		*/
/*	|		   + (pch, pcv)		|		*/
/*	|		      [cu, cv]		|		*/
/*	|					|		*/
/*	|					|		*/
/*	|					|		*/
/*	|					|		*/
/*	+---------------------------------------+  ----> v	*/
/*								*/
/*  where u, v can be any two of +/- {x, y, z}
/*								*/
/*  View is specified by three values each in +/- {1, 2, 3}:	*/
/*								*/
/*   1: x axis   2: y axis  3: z axis				*/
/*   ax, ay - horizonal, vertical axes for plot			*/
/*   az - depth axis						*/
/*   Positive values mean forward direction, or view from +oo	*/
/*   Negative values mean reverse direction, or view from -oo	*/
/*								*/
/*   Examples:  						*/
/*								*/
/*   ax = 1, ay = 2, az = 3:					*/
/*     right:  increasing x,  up:  increasing y,  closer: +z	*/
/*								*/
/*   ax = 1, ay = 3, az = -2;					*/
/*     right:  increasing x,  up:  increasing z,  closer: -z	*/
/*								*/
/*  Note that voxel coordinates use a Cartesian system that	*/
/*  corresponds directly to the space coordinates, unlike	*/
/*  2-D screen coordinates.					*/


import java.awt.*;
import java.io.*;
import java.lang.Math;

public class NBView extends Canvas
 { 
   public ViewGroup myGroup;

   public int	 pwd,	/* Canvas width	 (pixels)	*/
		 pht,	/* Canvas height (pixels)	*/
		 pch,	/* Center pixel (h)		*/
		 pcv,	/* Center pixel (v)		*/
          	 ax,	/* Horizontal axis +/- {1,2,3}	*/
		 ay,	/* Vertical axis   +/- {1,2,3}	*/
		 az,	/* Depth axis      +/- {1,2,3}	*/
		 ix,	/* Subscript for horiz. coord.	*/
		 iy,	/* Subscript for vert. coor.	*/
		 iz;	/* Subscript for depth.		*/

   ObjView obs[];	/* Objects plotted in this view	*/
   int     nobs;	/* Number of objects in obs	*/

   Image nbBuff;		/* The image buffer		*/
   public Graphics gcx;		/* Graphics context for nbBuff	*/

   Color bg = Color.black,	/* Standard background	*/
	 fg = Color.white;	/* Standard foreground	*/

   /*  Constructor	*/

   NBView(ViewGroup myg, int vx, int vy, int vz)
    { myGroup = myg;
      ax = vx;  ay = vy;  az = vz;
      ix = Math.abs(ax) - 1;  iy = Math.abs(ay) - 1;  iz = Math.abs(az) - 1;
      pwd = myGroup.pw[ix];  pht = myGroup.pw[iy];
      pch = myGroup.pc[ix];  pcv = myGroup.pc[iy];
      System.out.println("View:  "+ax+" "+ay+" "+az+" "+ix+" "+iy+" "+iz
					+" "+pwd+" "+pht+" "+pch+" "+pcv);
      resize(pwd, pht);		/* Make it the right size */
    }


   /*  Initialize - leave cleared image (black)	*/

   public void init()
    { setBackground(bg);  setForeground(fg);
      obs = new ObjView[myGroup.NOb];  nobs = 0;
      nbBuff = createImage(pwd, pht);
      gcx = nbBuff.getGraphics();
      clear();
    }


   /*  Add an object view	*/

   public ObjView loadObject(NBObj obj)
    { ObjView v = new ObjView(obj);  obs[nobs] = v;  nobs++;
      return v;
    }


   /* Clear buffer	*/

   public void clear()
    { gcx.setColor(bg);  gcx.fillRect(0, 0, pwd, pht); }


   /*  Update pixel coordinates in ov from current voxel coord's.	*/

   public void setcoords(ObjView ov)
    { NBObj ob = ov.myObj;  int c[] = ob.c;  Point p = ov.p;
      ov.p0.x = p.x;  ov.p0.y = p.y;     /* Save position */
      if ( ax > 0)  p.x = c[ix];  else  p.x = 2*pch - c[ix];
      if ( ay > 0)  p.y = 2*pcv - c[iy];  else  p.y = c[iy];
      if ( az < 0)  ov.depth = ob.x[iz];  else  ov.depth = -ob.x[iz];
    }


   /*  Update and draw all objects, back to front.	*/

   public void updateDrawAll()
    { for (int i = nobs-1;  i >= 0;  i-- )
       { ObjView ovw = obs[i];  NBObj ob = ovw.myObj;
	 setcoords(ovw);
       }
      depthSort();
      for (int i = nobs-1;  i >= 0;  i-- )
       { ObjView ovw = obs[i];  NBObj ob = ovw.myObj;
         ob.b.drawBody(gcx, ovw.p, ob.d0);
       }
    }

   /*  Update and move all objects, back to front, leaving a trail..	*/

   public void updateMoveAll()
    { for (int i = nobs-1;  i >= 0;  i-- )
       { ObjView ovw = obs[i];  NBObj ob = ovw.myObj;
	 setcoords(ovw);
       }
      depthSort();
      for (int i = nobs-1;  i >= 0;  i-- )
       { ObjView ovw = obs[i];  NBObj ob = ovw.myObj;
         ob.b.moveBody(gcx, ovw.p, ovw.p0, ob.d0, true);
       }
    }

   /* Recalculate pixel coords and draw all objects without resorting.	*/

   public void redrawAll()
    { for (int i = nobs-1;  i >= 0;  i-- )
       { ObjView ovw = obs[i];  NBObj ob = ovw.myObj;
	 setcoords(ovw);
         ob.b.drawBody(gcx, ovw.p, ob.d0);
       }
    }



   /* Sort objects so depth increases with index.  The algorithm is	*/
   /* intended for the situation that objects are nearly in order at	*/
   /* almost all times.  It would be inefficient otherwise.		*/

   public void depthSort()
    { int j;
      /* Inv:  obs[0:i-1] is ordered	*/
      for (int i = 1;  i < nobs;  i++)	/* Move obs[i] down to proper pos'n */
       {  ObjView ov = obs[i];  double dp = ov.depth;
          /* Inv:  ov fits at or below obs[j+1]		*/
	  /*       obs[0:j]+obs[j+2:i] is ordered	*/
	  for (j = i-1;  j >= 0 && dp < obs[j].depth;  j--)
	    obs[j+1] = obs[j];	/* Move item up	*/
	  obs[j+1] = ov;
       }
     }


   /* Respond to mouse clicks when recentering the view.	*/

   public boolean mouseDown(Event ev, int x, int y)
    { Semaphore pr = myGroup.prime;
      int mc[] = new int[3];  boolean result = false;
      // System.out.println("Mouse: "+x+" "+y+" "+pr.value);
      synchronized (pr)
       { if ( pr.value > 0 )		/* Calculate voxel displacement	*/
	  { int pc[] = myGroup.pc;	/* to new center.		*/
	    int xx = x - pc[ix];  if (ax < 0) xx = -xx;  mc[ix] = xx;
	    int yy = pc[iy] - y;  if (ay < 0) yy = -yy;  mc[iy] = yy;
	    mc[iz] = 0;
	    pr.value--;  result = true;
	  }
       }
      if (result) myGroup.voxCenter(mc);	/* Should be outside synch. */
      return result;
    }
      

   /* Flicker reduction with nbBuff	*/

   public void update(Graphics g)
    { paint(g); }

   public void repaint()
    { paint(getGraphics()); }

   public void paint(Graphics g)
    {  g.drawImage(nbBuff, 0, 0, this);
       // System.out.println("Painted view "+pwd+" "+pht);
    }

   public void destroy()
    { gcx.dispose(); }

 }
